"""Utilities for in the tests of the physbiblio modules.

This file is part of the physbiblio package.
"""
import sys
import traceback
import datetime
import os

if sys.version_info[0] < 3:
    import unittest2 as unittest
    from mock import patch
    from StringIO import StringIO

    USE_AUTOSPEC_CLASS = False
else:
    import unittest
    from unittest.mock import patch
    from io import StringIO

    USE_AUTOSPEC_CLASS = True

today_ymd = datetime.datetime.today().strftime("%y%m%d_%H%M%S")
tempLogFileName = "tests_%s.log" % today_ymd

try:
    from physbiblio.config import pbConfig

    pbConfig.overWritelogFileName = os.path.join(pbConfig.dataPath, tempLogFileName)
    pbConfig.prepareLogger("physbibliotestlog")
    pbConfig.reloadProfiles()
    from physbiblio.errors import pBErrorManager, pBLogger
    from physbiblio.database import PhysBiblioDB
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class SkipTestsSettingsClass:
    """Store settings for deciding the tests to skip"""

    def __init__(self):
        """Define non-gui skip settings"""
        self.default()

    def copy(self):
        """create a new instance, copy of the previous one"""
        new = SkipTestsSettingsClass()
        new.db = bool(self.db)
        new.gui = bool(self.gui)
        new.long = bool(self.long)
        new.oai = bool(self.oai)
        new.online = bool(self.online)
        return new

    def default(self):
        """Default settings"""
        self.db = False
        self.gui = False
        self.long = False
        self.oai = False
        self.online = False

    def __str__(self):
        """print current settings"""
        return "DB: %s\nGUI: %s\nlong: %s\nOnline: %s (OAI: %s)" % (
            self.db,
            self.gui,
            self.long,
            self.online,
            self.oai,
        )


skipTestsSettings = SkipTestsSettingsClass()

pbConfig.params["logFileName"] = tempLogFileName
logFileName = os.path.join(pbConfig.dataPath, pbConfig.params["logFileName"])

tempDBName = os.path.join(pbConfig.dataPath, "tests_%s.db" % today_ymd)
if os.path.exists(tempDBName):
    os.remove(tempDBName)
tempFDBName = os.path.join(pbConfig.dataPath, "tests_first_%s.db" % today_ymd)
if os.path.exists(tempFDBName):
    os.remove(tempFDBName)


class DBTestCase(unittest.TestCase):
    """define the class that will be used for database tests"""

    @classmethod
    def setUpClass(self):
        """Create a temporary database"""
        self.maxDiff = None
        self.pBDB = PhysBiblioDB(tempDBName, pBLogger)

    def tearDown(self):
        """Clean temporary database"""
        self.pBDB.undo(verbose=False)

    @patch("sys.stdout", new_callable=StringIO)
    def assert_stdout(self, function, expected_output, mock_stdout):
        """Catch and test stdout of the function"""
        pBErrorManager.tempHandler(sys.stdout, format="%(message)s")
        function()
        pBErrorManager.rmTempHandler()
        self.assertEqual(mock_stdout.getvalue(), expected_output)


def tearDownModule():
    """Clean temporary files at the end"""
    if os.path.exists(logFileName):
        os.remove(logFileName)
    if os.path.exists(tempDBName):
        os.remove(tempDBName)
    if os.path.exists(tempFDBName):
        os.remove(tempFDBName)



"""Module with the classes and functions that manage many generic
simple dialog windows of the PhysBiblio application.

This file is part of the physbiblio package.
"""
import sys
from PySide2.QtWidgets import (
    QDesktopWidget,
    QDialog,
    QFileDialog,
    QGridLayout,
    QInputDialog,
    QMessageBox,
    QTextEdit,
    QPushButton,
)

if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO


def askYesNo(message, title="Question"):
    """Uses `QMessageBox` to ask "Yes" or "No" for a given question.

    Parameters:
        message: the question to be displayed
        title: the window title

    Output:
        return True if the "Yes" button has been clicked,
            False otherwise
    """
    mbox = QMessageBox(QMessageBox.Question, title, message)
    yesButton = mbox.addButton(QMessageBox.Yes)
    noButton = mbox.addButton(QMessageBox.No)
    mbox.setDefaultButton(noButton)
    mbox.exec_()
    if mbox.clickedButton() == yesButton:
        return True
    else:
        return False


def infoMessage(message, title="Information"):
    """Uses `QMessageBox` to show a simple message.

    Parameters:
        message: the question to be displayed
        title: the window title
    """
    mbox = QMessageBox(QMessageBox.Information, title, message)
    mbox.exec_()


class LongInfoMessage(QDialog):
    """`infoMessage` version when a long text is expected"""

    def __init__(self, message, title="Information"):
        """Class constructor.

        Parameters:
            message: the question to be displayed
            title: the window title
        """
        QDialog.__init__(self)
        self.setWindowTitle(title)
        self.gridlayout = QGridLayout()
        self.textarea = QTextEdit(message)
        self.textarea.setReadOnly(True)
        self.gridlayout.addWidget(self.textarea, 0, 0, 1, 2)
        self.okbutton = QPushButton("OK", self)
        self.okbutton.clicked.connect(lambda: QDialog.close(self))
        self.gridlayout.addWidget(self.okbutton, 1, 1)
        self.setLayout(self.gridlayout)
        self.setGeometry(100, 100, 600, 400)
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())
        self.exec_()


def askGenericText(message, title, parent=None):
    """Uses `QInputDialog` to ask a text answer for a given question.

    Parameters:
        message: the question to be displayed
        title: the window title
        parent (optional, default None): the parent of the window

    Output:
        return a tuple containing the text as first element
        and True/False
            (depending if the user selected "Ok" or "Cancel")
            as the second element
    """
    dialog = QInputDialog(parent)
    dialog.setInputMode(QInputDialog.TextInput)
    dialog.setWindowTitle(title)
    dialog.setLabelText(message)
    out = dialog.exec_()
    return dialog.textValue(), out


def askFileName(parent=None, title="Filename to use:", filter="", dir=""):
    """Uses `QFileDialog` to ask the name of a single, existing file

    Parameters (all optional):
        parent (default None): the parent of the window
        title: the window title
        filter: the filter to be used when displaying files
        dir: the initial directory

    Output:
        return the filename
        or an empty string if the user selected "Cancel"
    """
    dialog = QFileDialog(parent, title, dir, filter)
    dialog.setFileMode(QFileDialog.ExistingFile)
    if dialog.exec_():
        fileNames = dialog.selectedFiles()
        try:
            return fileNames[0]
        except IndexError:
            return ""
    else:
        return ""


def askFileNames(parent=None, title="Filename to use:", filter="", dir=""):
    """Uses `QFileDialog` to ask the names of a set of existing files

    Parameters (all optional):
        parent (default None): the parent of the window
        title: the window title
        filter: the filter to be used when displaying files
        dir: the initial directory

    Output:
        return the filenames list
        or an empty list depending if the user selected "Ok" or "Cancel"
    """
    dialog = QFileDialog(parent, title, dir, filter)
    dialog.setFileMode(QFileDialog.ExistingFiles)
    dialog.setOption(QFileDialog.DontConfirmOverwrite, True)
    if dialog.exec_():
        fileNames = dialog.selectedFiles()
        return fileNames
    else:
        return []


def askSaveFileName(parent=None, title="Filename to use:", filter="", dir=""):
    """Uses `QFileDialog` to ask the names of a single file
    where something will be saved (the file may not exist)

    Parameters (all optional):
        parent (default None): the parent of the window
        title: the window title
        filter: the filter to be used when displaying files
        dir: the initial directory

    Output:
        return the filename
        or an empty string depending if the user selected "Ok" or "Cancel"
    """
    dialog = QFileDialog(parent, title, dir, filter)
    dialog.setFileMode(QFileDialog.AnyFile)
    dialog.setOption(QFileDialog.DontConfirmOverwrite, True)
    if dialog.exec_():
        fileNames = dialog.selectedFiles()
        try:
            return fileNames[0]
        except IndexError:
            return ""
    else:
        return ""


def askDirName(parent=None, title="Directory to use:", dir=""):
    """Uses `QFileDialog` to ask the names of a single directory

    Parameters (all optional):
        parent (default None): the parent of the window
        title: the window title
        dir: the initial directory

    Output:
        return the directory name
        or an empty string depending if the user selected "Ok" or "Cancel"
    """
    dialog = QFileDialog(parent, title, dir)
    dialog.setFileMode(QFileDialog.Directory)
    dialog.setOption(QFileDialog.ShowDirsOnly, True)
    if dialog.exec_():
        fileNames = dialog.selectedFiles()
        try:
            return fileNames[0]
        except IndexError:
            return ""
    else:
        return ""



"""Module that manages the command line calls of PhysBiblio.

This file is part of the physbiblio package.
"""
import sys
import datetime
import argparse

from PySide2.QtWidgets import QApplication

try:
    from physbiblio import __version__, __version_date__
    from physbiblio.config import pbConfig
    from physbiblio.errors import pBLogger
    from physbiblio.gui.errorManager import pBGUIErrorManager
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise


def call_clean(args):
    """Function used when the "clean" subcommand is called"""
    from physbiblio.database import pBDB

    pBDB.bibs.cleanBibtexs(startFrom=args.startFrom)
    pBDB.commit()


def call_cli(args):
    """Function used when the "cli" subcommand is called"""
    from physbiblio.cli import cli as physBiblioCLI

    physBiblioCLI()


def call_export(args):
    """Function used when the "export" subcommand is called"""
    from physbiblio.export import pBExport

    pBExport.exportAll(args.filename)


def call_tests(args):
    """Function used when the "test" subcommand is called"""
    from physbiblio.setuptests import skipTestsSettings

    if args.database:
        skipTestsSettings.db = True
    if args.gui:
        skipTestsSettings.gui = True
    if args.long:
        skipTestsSettings.long = True
    if args.online_oai:
        skipTestsSettings.oai = True
    if args.online:
        skipTestsSettings.online = True
        skipTestsSettings.oai = True

    from physbiblio.testLoader import PBScanningLoader

    if sys.version_info[0] < 3:
        import unittest2 as unittest
    else:
        import unittest
    suite = PBScanningLoader().discover("physbiblio")
    testRunner = unittest.runner.TextTestRunner()
    result = testRunner.run(suite)
    if not result.wasSuccessful():
        raise Exception("Some error occurred during tests")


def call_tex(args):
    """Function used when the "tex" subcommand is called"""
    from physbiblio.export import pBExport

    pBExport.exportForTexFile(
        args.texFiles,
        args.bibFile,
        overwrite=args.overwrite,
        autosave=args.save,
        updateExisting=args.updateExisting,
        removeUnused=args.removeUnused,
    )


def call_update(args):
    """Function used when the "update" subcommand is called"""
    from physbiblio.database import pBDB

    pBDB.bibs.searchOAIUpdates(startFrom=args.startFrom, force=args.force)
    pBDB.commit()


def call_oaiDates(date1, date2):
    """Wrapper for getDailyInfoFromOAI + commit of the changes"""
    from physbiblio.database import pBDB

    pBDB.bibs.getDailyInfoFromOAI(date1, date2)
    pBDB.commit()


dateLast = datetime.date.today().strftime("%Y-%m-%d")


def call_dates(args):
    """Use call_oaiDates to fetch updates between two dates
    specified in the command line
    """
    call_oaiDates(args.start, args.end)


def call_daily(args):
    """Use call_oaiDates to fetch daily updates"""
    call_oaiDates(
        (datetime.date.today() - datetime.timedelta(1)).strftime("%Y-%m-%d"), dateLast
    )


def call_weekly(args):
    """Use call_oaiDates to fetch weekly updates"""
    call_oaiDates(
        (datetime.date.today() - datetime.timedelta(7)).strftime("%Y-%m-%d"), dateLast
    )


def call_gui(args=None):
    """Function that runs the PhysBiblio GUI"""
    # these two imports must stay here,
    # so they start after the profile has been loaded properly:
    try:
        from physbiblio.database import pBDB
        import physbiblio.gui.mainWindow
    except ImportError:
        print("Could not find physbiblio and its modules!")
        raise
    try:
        app = QApplication(sys.argv)
        mainWin = physbiblio.gui.mainWindow.MainWindow()
        sys.excepthook = mainWin.errormessage.emit
        mainWin.show()
        mainWin.raise_()
        if pbConfig.params["openSinceLastUpdate"] != __version__:
            mainWin.recentChanges()
            pbConfig.globalDb.config.update("openSinceLastUpdate", __version__)
            pbConfig.globalDb.commit()
        sys.exit(app.exec_())
    except NameError:
        pBLogger.critical("NameError:", exc_info=True)
    except SystemExit:
        pBDB.closeDB()
        pBLogger.info("Closing main window...")


class NewProfileAction(argparse.Action):
    """Used to trigger a reload of the settings
    if a profile is specified as an argument
    """

    def __init__(self, option_strings, dest, **kwargs):
        super(NewProfileAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        prof = values[0]
        if prof in pbConfig.profiles.keys():
            pbConfig.reloadProfiles(prof)
        setattr(namespace, self.dest, values)


def setParser():
    """Contains the argparse definition of sub-commands and options"""
    parser = argparse.ArgumentParser(prog="PhysBiblio.exe")
    parser.add_argument(
        "-p",
        "--profile",
        action=NewProfileAction,
        nargs=1,
        choices=pbConfig.profiles.keys(),
        help="define the profile that must be used",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="PhysBiblio %s (%s)" % (__version__, __version_date__),
    )
    subparsers = parser.add_subparsers(help="sub-command help", dest="cmd")

    parser_clean = subparsers.add_parser(
        "clean", help="clean the entries in the database"
    )
    parser_clean.add_argument(
        "-s",
        "--startFrom",
        type=int,
        help="the index from which the cleaning should start",
        default=0,
    )
    parser_clean.set_defaults(func=call_clean)

    parser_cli = subparsers.add_parser(
        "cli", help="open the internal command line interface"
    )
    parser_cli.set_defaults(func=call_cli)

    parser_daily = subparsers.add_parser(
        "daily", help="fetch the daily updates from INSPIRE-HEP OAI"
    )
    parser_daily.set_defaults(func=call_daily)

    parser_dates = subparsers.add_parser(
        "dates", help="fetch the updates from INSPIRE-HEP OAI between the given dates"
    )
    parser_dates.add_argument(
        "start",
        metavar="startdate",
        help="the starting date (format as yyyy-mm-dd)",
        default=(datetime.date.today() - datetime.timedelta(1)).strftime("%Y-%m-%d"),
    )
    parser_dates.add_argument(
        "end",
        metavar="enddate",
        help="the ending date (format as yyyy-mm-dd)",
        default=dateLast,
    )
    parser_dates.set_defaults(func=call_dates)

    parser_export = subparsers.add_parser(
        "export", help="export all the entries in the database in a file"
    )
    parser_export.add_argument(
        "filename", help="the filename where to save the entries"
    )
    parser_export.set_defaults(func=call_export)

    parser_test = subparsers.add_parser("test", help="run the test suite")
    parser_test.add_argument(
        "-d",
        "--database",
        dest="database",
        action="store_true",
        help="do not perform database tests",
    )
    parser_test.add_argument(
        "-g", "--gui", dest="gui", action="store_true", help="do not perform gui tests"
    )
    parser_test.add_argument(
        "-l",
        "--long",
        dest="long",
        action="store_true",
        help="do not perform long tests",
    )
    parser_test.add_argument(
        "-o",
        "--online",
        dest="online",
        action="store_true",
        help="do not perform online tests",
    )
    parser_test.add_argument(
        "--online_oai",
        dest="online_oai",
        action="store_true",
        help="do not perform online tests with INSPIRE OAI",
    )
    parser_test.set_defaults(func=call_tests)

    parser_tex = subparsers.add_parser(
        "tex", help="read .tex file(s) and create a *.bib file with the cited bibtexs"
    )
    parser_tex.add_argument(
        "bibFile",
        metavar="bibfilename",
        help="the filename of the bib file where to write",
    )
    parser_tex.add_argument(
        "texFiles",
        metavar="texfilename",
        help="the filename of the tex file to read. "
        + "A folder or wildcards are admitted",
        nargs="+",
    )
    parser_tex.add_argument(
        "-o",
        "--overwrite",
        dest="overwrite",
        action="store_true",
        help="overwrite the bib file, if existing",
    )
    parser_tex.add_argument(
        "-r",
        "--removeUnused",
        dest="removeUnused",
        action="store_true",
        help="remove from the .bib file those elements "
        + "that are not more used in the tex",
    )
    parser_tex.add_argument(
        "-s",
        "--save",
        dest="save",
        action="store_false",
        help="save the changes in the database",
    )
    parser_tex.add_argument(
        "-u",
        "--updateExisting",
        dest="updateExisting",
        action="store_true",
        help="update the elements in the .bib file if they changed "
        + "in the database and remove duplicated entries",
    )
    parser_tex.set_defaults(func=call_tex)

    parser_update = subparsers.add_parser(
        "update", help="use INSPIRE to update the information in the database"
    )
    parser_update.add_argument(
        "-s",
        "--startFrom",
        type=int,
        help="the index from which the updating should start",
        default=0,
    )
    parser_update.add_argument(
        "-f", "--force", action="store_true", help="force the update"
    )
    parser_update.set_defaults(func=call_update)

    parser_weekly = subparsers.add_parser(
        "weekly", help="fetch the weekly updates from INSPIRE-HEP OAI"
    )
    parser_weekly.set_defaults(func=call_weekly)

    parser_gui = subparsers.add_parser("gui", help="open the gui")
    parser_gui.set_defaults(func=call_gui)

    return parser



#!/usr/bin/env python
"""Test file for the physbiblio.gui.errorManager module.

This file is part of the physbiblio package.
"""
import sys
import traceback
import os
import logging
from PySide2.QtCore import Qt
from PySide2.QtTest import QTest
from PySide2.QtWidgets import QMessageBox

if sys.version_info[0] < 3:
    import unittest2 as unittest
    from mock import patch, MagicMock
else:
    import unittest
    from unittest.mock import patch, MagicMock

try:
    from physbiblio.setuptests import *
    from physbiblio.gui.setuptests import *
    from physbiblio.errors import PBErrorManagerClass
    from physbiblio.gui.errorManager import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class TestGUIErrorManager(GUITestCase):
    """Test the functions in ErrorManager"""

    def test_ErrorStream(self):
        """Test functions in ErrorStream"""
        es = ErrorStream()
        self.assertIsInstance(es, StringIO)
        self.assertEqual(es.lastMBox, None)
        self.assertEqual(es.priority, 1)
        es.setPriority(2)
        self.assertEqual(es.priority, 2)
        mb = MagicMock()
        with patch(
            "physbiblio.gui.errorManager.QMessageBox", return_value=mb, autospec=True
        ) as _mb:
            es.write("sometext")
            _mb.assert_called_once_with(_mb.Information, "Information", "")
        mb.setText.assert_called_once_with("sometext")
        mb.setWindowTitle.assert_called_once_with("Error")
        mb.setIcon.assert_called_once_with(_mb.Critical)
        mb.exec_.assert_called_once_with()
        self.assertEqual(es.priority, 1)
        es.setPriority(0)
        mb = MagicMock()
        with patch(
            "physbiblio.gui.errorManager.QMessageBox", return_value=mb, autospec=True
        ) as _mb:
            es.write("some\ntext")
            _mb.assert_called_once_with(_mb.Information, "Information", "")
        mb.setText.assert_called_once_with("some<br>text")
        mb.setWindowTitle.assert_called_once_with("Information")
        mb.setIcon.assert_called_once_with(_mb.Information)
        mb.exec_.assert_called_once_with()
        mb = MagicMock()
        with patch(
            "physbiblio.gui.errorManager.QMessageBox", return_value=mb, autospec=True
        ) as _mb:
            es.write("some new text")
            _mb.assert_called_once_with(_mb.Information, "Information", "")
        mb.setText.assert_called_once_with("some new text")
        mb.setWindowTitle.assert_called_once_with("Warning")
        mb.setIcon.assert_called_once_with(_mb.Warning)
        mb.exec_.assert_called_once_with()
        with patch(
            "physbiblio.gui.errorManager.QMessageBox", return_value=mb, autospec=True
        ) as _mb:
            self.assertEqual(es.write(" "), None)
            self.assertEqual(_mb.call_count, 0)

    def test_PBErrorManagerClassGui(self):
        """Test functions in ErrorStream"""
        el = PBErrorManagerClassGui()
        self.assertIsInstance(el, PBErrorManagerClass)
        self.assertIsInstance(el.guiStream, ErrorStream)
        l = el.loggerPriority(25)
        self.assertEqual(el.guiStream.priority, 25)
        self.assertIsInstance(l, logging.Logger)

    def test_objects(self):
        """test that the module contains the appropriate objects"""
        import physbiblio.gui.errorManager as pgem

        self.assertIsInstance(pgem.pBGUIErrorManager, pgem.PBErrorManagerClassGui)
        self.assertIsInstance(pgem.pBGUIErrorManager, PBErrorManagerClass)
        self.assertIsInstance(pgem.pBGUILogger, logging.Logger)
        self.assertEqual(pgem.pBGUILogger, pgem.pBGUIErrorManager.logger)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""Test file for the physbiblio.gui.basicDialogs module.

This file is part of the physbiblio package.
"""
import sys
import traceback
import os
from PySide2.QtCore import Qt
from PySide2.QtTest import QTest
from PySide2.QtWidgets import QInputDialog

if sys.version_info[0] < 3:
    import unittest2 as unittest
    from mock import patch, MagicMock, call
else:
    import unittest
    from unittest.mock import patch, MagicMock, call

try:
    from physbiblio.setuptests import *
    from physbiblio.gui.setuptests import *
    from physbiblio.gui.basicDialogs import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class TestDialogWindows(GUITestCase):
    """Test the functions in DialogWindows"""

    def test_askYesNo(self):
        """Test askYesNo"""
        mb = MagicMock()
        mb.addButton.side_effect = ["yes", "no", "yes", "no"]
        mb.clickedButton.side_effect = ["yes", "no"]
        with patch(
            "physbiblio.gui.basicDialogs.QMessageBox", return_value=mb, autospec=True
        ) as _mb:
            self.assertTrue(askYesNo("mymessage"))
            _mb.assert_called_once_with(_mb.Question, "Question", "mymessage")
            mb.addButton.assert_has_calls([call(_mb.Yes), call(_mb.No)])
            mb.setDefaultButton.assert_called_once_with("no")
            mb.exec_.assert_called_once_with()
            mb.clickedButton.assert_called_once_with()
            _mb.reset_mock()
            self.assertFalse(askYesNo("mymessage", title="mytitle"))
            _mb.assert_called_once_with(_mb.Question, "mytitle", "mymessage")

    def test_infoMessage(self):
        """Test infoMessage"""
        mb = MagicMock()
        with patch(
            "physbiblio.gui.basicDialogs.QMessageBox", return_value=mb, autospec=True
        ) as _mb:
            infoMessage("mymessage")
            _mb.assert_called_once_with(_mb.Information, "Information", "mymessage")
            _mb.reset_mock()
            infoMessage("mymessage", title="mytitle")
            _mb.assert_called_once_with(_mb.Information, "mytitle", "mymessage")
            mb.exec_.assert_called_once_with()

    def test_LongInfoMessage(self):
        """Test LongInfoMessage"""
        with patch(
            "physbiblio.gui.basicDialogs.LongInfoMessage.exec_", autospec=True
        ) as _e:
            win = LongInfoMessage("mymessage")
            _e.assert_called_once_with()
        self.assertIsInstance(win, QDialog)
        self.assertEqual(win.windowTitle(), "Information")
        self.assertIsInstance(win.layout(), QGridLayout)
        self.assertEqual(win.layout(), win.gridlayout)
        self.assertIsInstance(win.layout().itemAtPosition(0, 0).widget(), QTextEdit)
        self.assertEqual(win.layout().itemAtPosition(0, 0).widget(), win.textarea)
        self.assertEqual(win.textarea.isReadOnly(), True)
        self.assertEqual(win.textarea.toPlainText(), "mymessage")
        self.assertIsInstance(win.layout().itemAtPosition(1, 1).widget(), QPushButton)
        self.assertEqual(win.layout().itemAtPosition(1, 1).widget(), win.okbutton)
        self.assertEqual(win.okbutton.text(), "OK")
        with patch("PySide2.QtWidgets.QDialog.close") as _c:
            QTest.mouseClick(win.okbutton, Qt.LeftButton)
            _c.assert_called_once_with(win)
        mb = MagicMock()
        with patch(
            "physbiblio.gui.basicDialogs.LongInfoMessage.exec_", autospec=True
        ) as _e:
            win = LongInfoMessage("mymessage", "mytitle")
        self.assertEqual(win.windowTitle(), "mytitle")

    def test_askGenericText(self):
        """Test askGenericText"""
        p = MagicMock()
        qid = MagicMock()
        qid.exec_.side_effect = [True, False]
        qid.textValue.side_effect = ["abc", "def"]
        with patch(
            "physbiblio.gui.basicDialogs.QInputDialog", return_value=qid, autospec=True
        ) as _qid:
            self.assertEqual(askGenericText("mymessage", "mytitle"), ("abc", True))
            qid.setInputMode.assert_called_once_with(_qid.TextInput)
            qid.setLabelText.assert_called_once_with("mymessage")
            qid.setWindowTitle.assert_called_once_with("mytitle")
            qid.exec_.assert_called_once_with()
            qid.textValue.assert_called_once_with()
            self.assertEqual(
                askGenericText("mymessage", "mytitle", parent=p), ("def", False)
            )

    def test_askFileName(self):
        """Test askFileName"""
        fd = MagicMock()
        fd.exec_.side_effect = [False, True, True]
        fd.selectedFiles.side_effect = [["abc", "def"], []]
        with patch(
            "physbiblio.gui.basicDialogs.QFileDialog", return_value=fd, autospec=True
        ) as _fd:
            self.assertEqual(
                askFileName(
                    parent=None, title="mytitle", dir="/tmp", filter="test: *.txt"
                ),
                "",
            )
            _fd.assert_called_once_with(None, "mytitle", "/tmp", "test: *.txt")
            fd.setFileMode.assert_called_once_with(_fd.ExistingFile)
            fd.exec_.assert_called_once_with()
            fd.selectedFiles.assert_not_called()
            _fd.reset_mock()
            self.assertEqual(askFileName(), "abc")
            _fd.assert_called_once_with(None, "Filename to use:", "", "")
            fd.selectedFiles.assert_called_once_with()
            self.assertEqual(askFileName(), "")

    def test_askFileNames(self):
        """Test askFileNames"""
        fd = MagicMock()
        fd.exec_.side_effect = [False, True]
        fd.selectedFiles.return_value = ["abc", "def"]
        with patch(
            "physbiblio.gui.basicDialogs.QFileDialog", return_value=fd, autospec=True
        ) as _fd:
            self.assertEqual(
                askFileNames(
                    parent=None, title="mytitle", dir="/tmp", filter="test: *.txt"
                ),
                [],
            )
            _fd.assert_called_once_with(None, "mytitle", "/tmp", "test: *.txt")
            fd.setFileMode.assert_called_once_with(_fd.ExistingFiles)
            fd.setOption.assert_called_once_with(_fd.DontConfirmOverwrite, True)
            fd.exec_.assert_called_once_with()
            fd.selectedFiles.assert_not_called()
            _fd.reset_mock()
            self.assertEqual(askFileNames(), ["abc", "def"])
            _fd.assert_called_once_with(None, "Filename to use:", "", "")
            fd.selectedFiles.assert_called_once_with()

    def test_askSaveFileName(self):
        """Test askSaveFileName"""
        fd = MagicMock()
        fd.exec_.side_effect = [False, True, True]
        fd.selectedFiles.side_effect = [["abc", "def"], []]
        with patch(
            "physbiblio.gui.basicDialogs.QFileDialog", return_value=fd, autospec=True
        ) as _fd:
            self.assertEqual(
                askSaveFileName(
                    parent=None, title="mytitle", dir="/tmp", filter="test: *.txt"
                ),
                "",
            )
            _fd.assert_called_once_with(None, "mytitle", "/tmp", "test: *.txt")
            fd.setFileMode.assert_called_once_with(_fd.AnyFile)
            fd.setOption.assert_called_once_with(_fd.DontConfirmOverwrite, True)
            fd.exec_.assert_called_once_with()
            fd.selectedFiles.assert_not_called()
            _fd.reset_mock()
            self.assertEqual(askSaveFileName(), "abc")
            _fd.assert_called_once_with(None, "Filename to use:", "", "")
            fd.selectedFiles.assert_called_once_with()
            self.assertEqual(askSaveFileName(), "")

    def test_askDirName(self):
        """Test askDirName"""
        fd = MagicMock()
        fd.exec_.side_effect = [False, True, True]
        fd.selectedFiles.side_effect = [["abc", "def"], []]
        with patch(
            "physbiblio.gui.basicDialogs.QFileDialog", return_value=fd, autospec=True
        ) as _fd:
            self.assertEqual(askDirName(parent=None, title="mytitle", dir="/tmp"), "")
            _fd.assert_called_once_with(None, "mytitle", "/tmp")
            fd.setFileMode.assert_called_once_with(_fd.Directory)
            fd.setOption.assert_called_once_with(_fd.ShowDirsOnly, True)
            fd.exec_.assert_called_once_with()
            fd.selectedFiles.assert_not_called()
            _fd.reset_mock()
            self.assertEqual(askDirName(), "abc")
            _fd.assert_called_once_with(None, "Directory to use:", "")
            fd.selectedFiles.assert_called_once_with()
            self.assertEqual(askDirName(), "")


if __name__ == "__main__":
    unittest.main()



"""Module that opens the external links.

This file is part of the physbiblio package.
"""
import subprocess
import traceback

try:
    from physbiblio.errors import pBLogger
    from physbiblio.config import pbConfig
    from physbiblio.database import pBDB
    from physbiblio.pdf import pBPDF
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class ViewEntry:
    """Contains methods to print or open a web link to the entry"""

    inspireRecord = pbConfig.inspireRecord
    inspireSearch = pbConfig.inspireSearchBase + "?p=find+"

    def __init__(self):
        """Init the class, storing the name of
        the external web application
        and the base strings to build some links
        """
        self.webApp = pbConfig.params["webApplication"]
        self.inspireRecord = pbConfig.inspireRecord
        self.inspireSearch = pbConfig.inspireSearchBase + "?p=find+"

    def getLink(self, key, arg="arxiv", fileArg=None):
        """Uses database information to construct and
        print the web link, or the pdf module to open a pdf

        Parameters:
            key: the bibtex identifier as stored in the database.
                It may be a list (calls recursively itself)
            arg: the string defining the type of link/action
                ("arxiv", "doi", "inspire", "file")
            fileArg: additional argument for the PDF module,
                used only if arg == "file"

        Output:
            link: the string of the web link
                if arg in ("arxiv", "doi", "inspire"),
                True if arg == "file",
                False if arg is invalid
        """
        if isinstance(key, list):
            links = []
            for k in key:
                if arg is not "file":
                    links.append(self.getLink(k, arg, fileArg))
                else:
                    links.append(fileArg)
            return links
        ads = pBDB.bibs.getField(key, "ads")
        arxiv = pBDB.bibs.getField(key, "arxiv")
        doi = pBDB.bibs.getField(key, "doi")
        inspire = pBDB.bibs.getField(key, "inspire")
        if arg is "ads" and ads:
            link = pBDB.bibs.getAdsUrl(key)
        elif arg is "arxiv" and arxiv:
            link = pBDB.bibs.getArxivUrl(key, "abs")
        elif arg is "doi" and doi:
            link = pBDB.bibs.getDoiUrl(key)
        elif arg is "inspire" and inspire:
            link = self.inspireRecord + inspire
        elif arg is "inspire" and arxiv:
            link = self.inspireSearch + arxiv
        elif arg is "inspire":
            link = self.inspireSearch + key
        elif arg is "file":
            pBPDF.openFile(key, fileArg)
            return fileArg
        else:
            pBLogger.warning(
                "Invalid selection or missing information.\n"
                + "Use one of (arxiv|doi|inspire|file) and "
                + "check that all the information is available "
                + "for entry '%s'." % key
            )
            return False

        return link

    def openLink(self, key, arg="arxiv", fileArg=None):
        """Uses the getLink method to generate the web link
        and opens it in an external application

        Parameters:
            key, arg, fileArg as in the getLink method
        """
        if isinstance(key, list):
            for k in key:
                self.openLink(k, arg, fileArg)
        else:
            if arg is "file":
                self.getLink(key, arg=arg, fileArg=fileArg)
                return
            elif arg is "link":
                link = key
            else:
                link = self.getLink(key, arg=arg, fileArg=fileArg)
            if link:
                if self.webApp != "":
                    pBLogger.info("Opening '%s'..." % link)
                    try:
                        subprocess.Popen(
                            [self.webApp, link],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                        )
                    except OSError:
                        pBLogger.warning("Opening link for '%s' failed!" % key)
            else:
                pBLogger.warning(
                    "Impossible to get the '%s'" % arg + " link for entry '%s'" % key
                )


pBView = ViewEntry()



"""Static definitions of the structure (column names and content)
for the tables of the database.

This file is part of the physbiblio package.
"""
profilesSettingsTable = [
    ["name", "text", "primary key not null"],
    ["databasefile", "text", "not null"],
    ["description", "text", ""],
    ["oldCfg", "text", ""],
    ["isDefault", "integer", "default 0"],
    ["ord", "integer", "default 100"],
    ["CONSTRAINT", "unique_databasefile", "UNIQUE (databasefile)"],
]
searchesTable = [
    ["idS", "integer", "primary key"],
    ["name", "text", ""],
    ["count", "integer", "default 0 not null"],
    ["searchDict", "text", "not null"],
    ["limitNum", "integer", "default 50"],
    ["offsetNum", "integer", "default 0"],
    ["replaceFields", "text", ""],
    ["manual", "integer", "default 0"],
    ["isReplace", "integer", "default 0"],
]
searchesTableDescriptions = {
    "idS": "Unique identifier",
    "name": "Custom name of the search/replace",
    "count": "Order of the entry in the cronology",
    "searchDict": "Dictionary of fields to be passed to fetchByDict",
    "limit": "The max number of results in the search",
    "offset": "The offset in the search",
    "replaceFields": "List of fields used in replacement",
    "manual": "Manually saved",
    "isReplace": "(T/F) A replacement or a simple search",
}

tableFields = {}
tableFields["entries"] = [
    ["bibkey", "text", "primary key not null collate nocase"],
    ["inspire", "text", ""],
    ["arxiv", "text", ""],
    ["ads", "text", ""],
    ["scholar", "text", ""],
    ["doi", "text", ""],
    ["isbn", "text", ""],
    ["year", "integer", ""],
    ["link", "text", ""],
    ["comments", "text", ""],
    ["old_keys", "text", ""],
    ["crossref", "text", ""],
    ["bibtex", "text", "not null"],
    ["firstdate", "text", "not null"],
    ["pubdate", "text", ""],
    ["exp_paper", "integer", "default 0"],
    ["lecture", "integer", "default 0"],
    ["phd_thesis", "integer", "default 0"],
    ["review", "integer", "default 0"],
    ["proceeding", "integer", "default 0"],
    ["book", "integer", "default 0"],
    ["noUpdate", "integer", "default 0"],
    ["marks", "text", ""],
    ["abstract", "text", ""],
    ["bibdict", "text", ""],
]
tableFields["categories"] = [
    ["idCat", "integer", "primary key"],
    ["name", "text", "not null"],
    ["description", "text", "not null"],
    ["parentCat", "integer", "default 0"],
    ["comments", "text", "default ''"],
    ["ord", "integer", "default 0"],
]
tableFields["experiments"] = [
    ["idExp", "integer", "primary key"],
    ["name", "text", "not null"],
    ["comments", "text", "not null"],
    ["homepage", "text", ""],
    ["inspire", "text", ""],
]
tableFields["expCats"] = [
    ["idExC", "integer", "primary key"],
    ["idExp", "integer", "not null"],
    ["idCat", "integer", "not null"],
]
tableFields["entryCats"] = [
    ["idEnC", "integer", "primary key"],
    ["bibkey", "text", "not null"],
    ["idCat", "integer", "not null"],
]
tableFields["entryExps"] = [
    ["idEnEx", "integer", "primary key"],
    ["bibkey", "text", "not null"],
    ["idExp", "integer", "not null"],
]
tableFields["settings"] = [
    ["name", "text", "primary key not null"],
    ["value", "text", "default ''"],
]
fieldsDescriptions = {}
fieldsDescriptions["entries"] = {
    "bibkey": "Unique bibtex key that identifies the bibliographic element",
    "inspire": "INSPIRE-HEP ID of the record",
    "arxiv": "arXiv ID of the record",
    "ads": "NASA ADS ID of the record",
    "scholar": "Google Scholar ID of the record",
    "doi": "DOI of the record",
    "isbn": "ISBN code of the record",
    "year": "Year of publication",
    "link": "Web link to article or additional material",
    "comments": "Comments",
    "old_keys": "Previous bibtex keys of the record",
    "crossref": "Bibtex crossref reference",
    "bibtex": "Bibtex entry",
    "firstdate": "Date of first appearance",
    "pubdate": "Date of publication",
    "exp_paper": "(T/F) The entry is a collaboration paper of some experiment",
    "lecture": "(T/F) The entry is a lecture",
    "phd_thesis": "(T/F) The entry is a PhD thesis",
    "review": "(T/F) The entry is a review",
    "proceeding": "(T/F) The entry is a proceeding",
    "book": "(T/F) The entry is a book",
    "noUpdate": "(T/F) The entry has been created/modified by the user "
    + "and/or should not be updated",
    "marks": "Mark the record",
    "abstract": "Abstract of the record",
    "bibdict": "Dictionary with fields of the bibtex entry from bibtexparser",
}
fieldsDescriptions["categories"] = {
    "idCat": "Unique ID that identifies the category",
    "name": "Name of the category",
    "description": "Description of the category",
    "parentCat": "Parent category",
    "comments": "Comments",
    "ord": "Ordering when plotting (not yet implemented)",
}
fieldsDescriptions["experiments"] = {
    "idExp": "Unique ID that identifies the experiment",
    "name": "Name of the experiment",
    "comments": "Description or comments",
    "homepage": "Web link to the experiment homepage",
    "inspire": "INSPIRE-HEP ID of the experiment record",
}
fieldsDescriptions["expCats"] = {
    "idExC": "Unique identifier",
    "idExp": "Corresponding experiment ID",
    "idCat": "Corresponding category ID",
}
fieldsDescriptions["entryCats"] = {
    "idEnC": "Unique identifier",
    "bibkey": "Corresponding bibtex key",
    "idCat": "Corresponding category ID",
}
fieldsDescriptions["entryExps"] = {
    "idEnEx": "Unique identifier",
    "bibkey": "Corresponding bibtex key",
    "idExp": "Corresponding experiment ID",
}
fieldsDescriptions["settings"] = {
    "id": "Unique identifier",
    "name": "name of the setting",
    "value": "value of the setting",
}



"""Tests for the `physbiblio.webimport` package.

This file is part of the physbiblio package.
"""
from physbiblio import __author__, __email__, __version__, __version_date__

__all__ = ["test_webimport"]



"""Utilities for in the tests of the physbiblio modules.

This file is part of the physbiblio package.
"""
import sys
import traceback
from PySide2.QtWidgets import QApplication, QFileDialog

if sys.version_info[0] < 3:
    import unittest2 as unittest
    from mock import MagicMock, patch
    from StringIO import StringIO
else:
    import unittest
    from unittest.mock import MagicMock, patch
    from io import StringIO

try:
    from physbiblio.gui.mainWindow import MainWindow
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())

globalQApp = QApplication()
globalQApp.exec_ = MagicMock(return_value=0)


def fakeExec(x, string, out):
    """Simulate the selection of some files and
    return True/False as if confirmed/canceled
    """
    QFileDialog.selectFile(x, string)
    return out


class GUITestCase(unittest.TestCase):
    """Class that manages GUI tests"""

    @classmethod
    def setUpClass(self):
        """Assign a temporary QApplication to the class instance"""
        self.maxDiff = None
        self.qapp = globalQApp

    @classmethod
    def tearDownClass(self):
        """Remove a temporary QApplication from the class instance"""
        del self.qapp

    def assertGeometry(self, obj, x, y, w, h):
        """Test the geometry of an object

        Parameters:
            obj: the object to be tested
            x: the expected x position
            y: the expected y position
            w: the expected width
            h: the expected height
        """
        geom = obj.geometry()
        self.assertEqual(geom.x(), x)
        self.assertEqual(geom.y(), y)
        self.assertEqual(geom.width(), w)
        self.assertEqual(geom.height(), h)


class GUIwMainWTestCase(unittest.TestCase):
    """Class that manages GUI tests which need a testing instance
    of the MainWindow class
    """

    @classmethod
    def setUpClass(self):
        """Call the parent method and instantiate a testing MainWindow"""
        super(GUIwMainWTestCase, self).setUpClass()
        self.mainW = MainWindow(testing=True)



"""The package that contains all the modules used by the PhysBiblio,
including the `gui` subpackage.

This file is part of the physbiblio package.
"""
__author__ = "Stefano Gariazzo"
__email__ = "stefano.gariazzo@gmail.com"

__version__ = "1.4.1"
__version_date__ = "14/10/2019"

__all__ = [
    "bibtexWriter",
    "cli",
    "config",
    "databaseCore",
    "database",
    "errors",
    "export",
    "inspireStats",
    "parseAccents",
    "pdf",
    "tablesDef",
    "view",
    "tests",
    "gui",
    "webimport",
]

__recent_changes__ = """<br>
* disable already useless stop button when a thread is finished;<br>
* attempt to fix "QThread: Destroyed while thread is still running";<br>
* fixing bugs with progress reporting in GUI for threaded process (and some functions improved meanwhile).<br>
"""



#!/usr/bin/env python
"""Test file for the physbiblio.gui.threadElements module.

This file is part of the physbiblio package.
"""
import sys
import traceback
import os
from PySide2.QtCore import Qt, Signal
from PySide2.QtTest import QTest
from PySide2.QtWidgets import QWidget

if sys.version_info[0] < 3:
    import unittest2 as unittest
    from mock import call, patch, MagicMock
    from urllib2 import URLError
    from Queue import Queue
else:
    import unittest
    from unittest.mock import call, patch, MagicMock
    from urllib.request import URLError
    from queue import Queue

try:
    from physbiblio import __version__
    from physbiblio.setuptests import *
    from physbiblio.database import pBDB
    from physbiblio.inspireStats import pBStats
    from physbiblio.gui.setuptests import *
    from physbiblio.gui.threadElements import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_checkUpdated(GUITestCase):
    """Test the functions in threadElements.Thread_checkUpdated"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        thr = Thread_checkUpdated(p)
        self.assertIsInstance(thr, PBThread)
        self.assertIsInstance(thr.result, Signal)
        self.assertEqual(thr.parent(), p)
        self.assertRaises(NotImplementedError, thr.setStopFlag)

    def test_run(self):
        """test run"""

        def fake_urlerror(x, y):
            raise URLError("no connection")

        p = QWidget()
        h1 = MagicMock()
        thr = Thread_checkUpdated(p)
        thr.result.connect(h1)
        with patch(
            "physbiblio.gui.threadElements.check_outdated",
            return_value=(False, __version__),
            autospec=True,
        ) as _cho:
            thr.run()
            _cho.assert_called_once_with("physbiblio", __version__)
        h1.assert_called_once_with(False, __version__)
        h1.reset_mock()
        with patch(
            "physbiblio.gui.threadElements.check_outdated",
            return_value=(),
            autospec=True,
        ) as _cho, patch("logging.Logger.warning") as _w:
            thr.run()
            _cho.assert_called_once_with("physbiblio", __version__)
            _w.assert_called_once_with(
                "Error when executing check_outdated. "
                + "Maybe you are using a developing version",
                exc_info=True,
            )
        h1.assert_not_called()
        with patch(
            "physbiblio.gui.threadElements.check_outdated", new=fake_urlerror
        ) as _cho, patch("logging.Logger.warning") as _w:
            thr.run()
            _w.assert_called_once_with(
                "Error when trying to check new versions. Are you offline?",
                exc_info=True,
            )
        with patch(
            "physbiblio.gui.threadElements.check_outdated",
            side_effect=ConnectionError,
            autospec=True,
        ) as _cho, patch("logging.Logger.warning") as _w:
            thr.run()
            _w.assert_called_once_with(
                "Error when trying to check new versions. Are you offline?",
                exc_info=True,
            )


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_updateAllBibtexs(GUITestCase):
    """Test the functions in threadElements.Thread_updateAllBibtexs"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_updateAllBibtexs(ws, 123, p, [[]], True, "r", pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.startFrom, 123)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.useEntries, [[]])
        self.assertEqual(thr.force, True)
        self.assertEqual(thr.reloadAll, "r")
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_updateAllBibtexs(
            ws, 123, p, [[]], True, True, pbMax="m", pbVal="v"
        )
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.searchOAIUpdates", autospec=True
        ) as _sou, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _sou.assert_called_once_with(
                pBDB.bibs,
                123,
                entries=[[]],
                force=True,
                reloadAll=True,
                pbMax="m",
                pbVal="v",
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

    def test_setStopFlag(self):
        """test setStopFlag"""
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_updateAllBibtexs(ws, 0)
        pBDB.bibs.runningOAIUpdates = True
        self.assertTrue(pBDB.bibs.runningOAIUpdates)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.runningOAIUpdates)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_replace(GUITestCase):
    """Test the functions in threadElements.Thread_updateAllBibtexs"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_replace(
            ws, "a", ["b"], "1", ["2"], parent=p, regex=True, pbMax="m", pbVal="v"
        )
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.fiOld, "a")
        self.assertEqual(thr.fiNew, ["b"])
        self.assertEqual(thr.old, "1")
        self.assertEqual(thr.new, ["2"])
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.regex, True)
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_replace(ws, "a", ["b"], "1", ["2"], p, True, pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        pBDB.bibs.lastFetched = ["e", "f"]
        with patch(
            "physbiblio.database.Entries.replace",
            return_value=("r1", "r2", "r3"),
            autospec=True,
        ) as _rep, patch("time.sleep") as _sl, patch(
            "physbiblio.database.Entries.fetchFromLast",
            return_value=pBDB.bibs,
            autospec=True,
        ) as _ffl, patch(
            "physbiblio.database.Entries.fetchCursor",
            return_value="curs",
            autospec=True,
        ) as _fc, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _rep.assert_called_once_with(
                pBDB.bibs,
                "a",
                ["b"],
                "1",
                ["2"],
                entries="curs",
                lenEntries=2,
                regex=True,
                pbMax="m",
                pbVal="v",
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
            _ffl.assert_has_calls([call(pBDB.bibs, doFetch=False), call(pBDB.bibs)])
            _fc.assert_called_once_with(pBDB.bibs)
            self.assertEqual(p.replaceResults, ("r1", "r2", "r3"))

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_replace(ws, "a", ["b"], "1", ["2"], p)
        pBDB.bibs.runningReplace = True
        self.assertTrue(pBDB.bibs.runningReplace)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.runningReplace)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_updateInspireInfo(GUITestCase):
    """Test the functions in threadElements.Thread_updateInspireInfo"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_updateInspireInfo(ws, "Gariazzo:2015rra", 1385583, p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.bibkey, "Gariazzo:2015rra")
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.inspireID, 1385583)
        self.assertRaises(NotImplementedError, thr.setStopFlag)

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_updateInspireInfo(ws, "Gariazzo:2015rra", 1385583, p)
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.updateInfoFromOAI", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBDB.bibs, 1385583, originalKey="Gariazzo:2015rra", verbose=1
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
        thr = Thread_updateInspireInfo(ws, "Gariazzo:2015rra", None, p)
        with patch(
            "physbiblio.database.Entries.updateInfoFromOAI", autospec=True
        ) as _fun, patch(
            "physbiblio.database.Entries.updateInspireID",
            return_value=1385,
            autospec=True,
        ) as _uiid, patch(
            "time.sleep"
        ) as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _uiid.assert_called_once_with(pBDB.bibs, "Gariazzo:2015rra")
            _fun.assert_called_once_with(pBDB.bibs, 1385, originalKey=None, verbose=1)
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

        ws = WriteStream(q)
        thr = Thread_updateInspireInfo(
            ws, ["Gariazzo:2015rra", "Gariazzo:2018mwd"], [1385583, None], p
        )
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.updateInfoFromOAI", autospec=True
        ) as _fun, patch(
            "physbiblio.database.Entries.updateInspireID",
            return_value=1385,
            autospec=True,
        ) as _uiid, patch(
            "time.sleep"
        ) as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_has_calls(
                [
                    call(pBDB.bibs, 1385583, originalKey="Gariazzo:2015rra", verbose=1),
                    call(pBDB.bibs, 1385, originalKey=None, verbose=1),
                ]
            )
            _uiid.assert_called_once_with(pBDB.bibs, "Gariazzo:2018mwd")
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_downloadArxiv(GUITestCase):
    """Test the functions in threadElements.Thread_downloadArxiv"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        thr = Thread_downloadArxiv("Gariazzo:2015rra", p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.bibkey, "Gariazzo:2015rra")
        self.assertRaises(NotImplementedError, thr.setStopFlag)

    def test_run(self):
        """test run"""
        p = QWidget()
        thr = Thread_downloadArxiv("Gariazzo:2015rra", p)
        with patch("physbiblio.pdf.LocalPDF.downloadArxiv", autospec=True) as _fun:
            thr.run()
            _fun.assert_called_once_with(pBPDF, "Gariazzo:2015rra")


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_processLatex(GUITestCase):
    """Test the functions in threadElements.Thread_processLatex"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        func = MagicMock()
        thr = Thread_processLatex(func, p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.func, func)
        self.assertIsInstance(thr.passData, Signal)
        self.assertRaises(NotImplementedError, thr.setStopFlag)

    def test_run(self):
        """test run"""
        p = QWidget()
        func = MagicMock(return_value=(["a", "b"], "text"))
        thr = Thread_processLatex(func, p)
        with MagicMock() as h1:
            thr.passData.connect(h1)
            thr.run()
            func.assert_called_once_with()
            h1.assert_called_once_with(["a", "b"], "text")


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_authorStats(GUITestCase):
    """Test the functions in threadElements.Thread_authorStats"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_authorStats(ws, "Stefano.Gariazzo.1", p, pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.authorName, "Stefano.Gariazzo.1")
        self.assertEqual(thr.receiver, ws)
        self.assertFalse(p.lastAuthorStats)
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")
        self.assertRaises(
            Exception, lambda: Thread_authorStats(ws, "Stefano.Gariazzo.1", None)
        )

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_authorStats(ws, "Stefano.Gariazzo.1", p, pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.inspireStats.InspireStatsLoader.authorStats",
            return_value={"a": "b"},
            autospec=True,
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBStats, "Stefano.Gariazzo.1", pbMax="m", pbVal="v"
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
            self.assertEqual(p.lastAuthorStats, {"a": "b"})

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_authorStats(ws, "Stefano.Gariazzo.1", p)
        pBStats.runningAuthorStats = True
        self.assertTrue(pBStats.runningAuthorStats)
        thr.setStopFlag()
        self.assertFalse(pBStats.runningAuthorStats)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_paperStats(GUITestCase):
    """Test the functions in threadElements.Thread_paperStats"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_paperStats(ws, 1385583, p, pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.inspireId, 1385583)
        self.assertEqual(thr.receiver, ws)
        self.assertFalse(p.lastPaperStats)
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")
        self.assertRaises(Exception, lambda: Thread_paperStats(ws, 1385583, None))

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_paperStats(ws, 1385583, p, pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.inspireStats.InspireStatsLoader.paperStats",
            return_value={"a": "b"},
            autospec=True,
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(pBStats, 1385583, pbMax="m", pbVal="v")
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
            self.assertEqual(p.lastPaperStats, {"a": "b"})

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_paperStats(ws, [1385583], p)
        pBStats.runningAuthorStats = True
        self.assertTrue(pBStats.runningPaperStats)
        thr.setStopFlag()
        self.assertFalse(pBStats.runningPaperStats)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_loadAndInsert(GUITestCase):
    """Test the functions in threadElements.Thread_loadAndInsert"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_loadAndInsert(ws, "Gariazzo:2015rra", p, pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.content, "Gariazzo:2015rra")
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")
        self.assertRaises(Exception, lambda: Thread_loadAndInsert(ws, 1385583, None))

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_loadAndInsert(ws, "Gariazzo:2015rra", p, pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        pBDB.bibs.lastInserted = ["def"]
        with patch(
            "physbiblio.database.Entries.loadAndInsert",
            side_effect=[True, False],
            autospec=True,
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBDB.bibs, "Gariazzo:2015rra", pbMax="m", pbVal="v"
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
            self.assertEqual(p.loadedAndInserted, ["def"])
            thr.run()
            self.assertEqual(p.loadedAndInserted, [])

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_loadAndInsert(ws, "Gariazzo:2015rra", p)
        pBDB.bibs.runningLoadAndInsert = True
        self.assertTrue(pBDB.bibs.runningLoadAndInsert)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.runningLoadAndInsert)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_cleanAllBibtexs(GUITestCase):
    """Test the functions in threadElements.Thread_cleanAllBibtexs"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanAllBibtexs(ws, 123, p, [[]], pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.startFrom, 123)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.useEntries, [[]])
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanAllBibtexs(ws, 123, p, [[]], pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.cleanBibtexs", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBDB.bibs, 123, entries=[[]], pbMax="m", pbVal="v"
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanAllBibtexs(ws, 123, p, [[]])
        pBDB.bibs.runningCleanBibtexs = True
        self.assertTrue(pBDB.bibs.runningCleanBibtexs)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.runningCleanBibtexs)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_findBadBibtexs(GUITestCase):
    """Test the functions in threadElements.Thread_findBadBibtexs"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_findBadBibtexs(ws, 123, p, [[]], pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.startFrom, 123)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.useEntries, [[]])
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")
        self.assertRaises(Exception, lambda: Thread_loadAndInsert(ws, 1385583, None))

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_findBadBibtexs(ws, 123, p, [[]], pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.findCorruptedBibtexs", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBDB.bibs, 123, entries=[[]], pbMax="m", pbVal="v"
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_findBadBibtexs(ws, 123, p, [[]])
        pBDB.bibs.runningFindBadBibtexs = True
        self.assertTrue(pBDB.bibs.runningFindBadBibtexs)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.runningFindBadBibtexs)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_importFromBib(GUITestCase):
    """Test the functions in threadElements.Thread_importFromBib"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_importFromBib(ws, "tmp.bib", False, p, pbMax="m", pbVal="v")
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.bibFile, "tmp.bib")
        self.assertFalse(thr.complete)
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_importFromBib(ws, "tmp.bib", False, p, pbMax="m", pbVal="v")
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.importFromBib", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBDB.bibs, "tmp.bib", False, pbMax="m", pbVal="v"
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_importFromBib(ws, "tmp.bib", False, p)
        pBDB.bibs.importFromBibFlag = True
        self.assertTrue(pBDB.bibs.importFromBibFlag)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.importFromBibFlag)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_exportTexBib(GUITestCase):
    """Test the functions in threadElements.Thread_exportTexBib"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_exportTexBib(ws, ["main.tex"], "biblio.bib", p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.outFName, "biblio.bib")
        self.assertEqual(thr.texFiles, ["main.tex"])
        self.assertFalse(thr.updateExisting)
        self.assertFalse(thr.removeUnused)
        thr = Thread_exportTexBib(
            ws,
            ["main.tex"],
            "biblio.bib",
            removeUnused=True,
            updateExisting=True,
            parent=p,
        )
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.outFName, "biblio.bib")
        self.assertEqual(thr.texFiles, ["main.tex"])
        self.assertTrue(thr.updateExisting)
        self.assertTrue(thr.removeUnused)

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_exportTexBib(ws, ["main.tex"], "biblio.bib", p)
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.export.PBExport.exportForTexFile", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBExport,
                ["main.tex"],
                "biblio.bib",
                removeUnused=False,
                updateExisting=False,
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
        ws = WriteStream(q)
        thr = Thread_exportTexBib(ws, ["main.tex"], "biblio.bib", p, True, True)
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.export.PBExport.exportForTexFile", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBExport,
                ["main.tex"],
                "biblio.bib",
                removeUnused=True,
                updateExisting=True,
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_exportTexBib(ws, ["main.tex"], "biblio.bib", p)
        pBExport.exportForTexFlag = True
        self.assertTrue(pBExport.exportForTexFlag)
        thr.setStopFlag()
        self.assertFalse(pBExport.exportForTexFlag)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_cleanSpare(GUITestCase):
    """Test the functions in threadElements.Thread_cleanSpare"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanSpare(ws, p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertRaises(NotImplementedError, thr.setStopFlag)

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanSpare(ws, p)
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Utilities.cleanSpareEntries", autospec=True
        ) as _fun, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(pBDB.utils)
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_cleanSparePDF(GUITestCase):
    """Test the functions in threadElements.Thread_cleanSparePDF"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanSparePDF(ws, p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertRaises(NotImplementedError, thr.setStopFlag)

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_cleanSparePDF(ws, p)
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.pdf.LocalPDF.removeSparePDFFolders", autospec=True
        ) as _fun, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(pBPDF)
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_fieldsArxiv(GUITestCase):
    """Test the functions in threadElements.Thread_fieldsArxiv"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_fieldsArxiv(
            ws, ["a", "b"], ["abstract", "arxiv"], p, pbMax="m", pbVal="v"
        )
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.entries, ["a", "b"])
        self.assertEqual(thr.fields, ["abstract", "arxiv"])
        self.assertEqual(thr.pbMax, "m")
        self.assertEqual(thr.pbVal, "v")

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_fieldsArxiv(
            ws, ["a", "b"], ["abstract", "arxiv"], p, pbMax="m", pbVal="v"
        )
        self.assertTrue(ws.running)
        with patch(
            "physbiblio.database.Entries.getFieldsFromArxiv", autospec=True
        ) as _fun, patch("time.sleep") as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            _fun.assert_called_once_with(
                pBDB.bibs, ["a", "b"], ["abstract", "arxiv"], pbMax="m", pbVal="v"
            )
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_fieldsArxiv(ws, ["a", "b"], ["abstract", "arxiv"], p)
        pBDB.bibs.getArxivFieldsFlag = True
        self.assertTrue(pBDB.bibs.getArxivFieldsFlag)
        thr.setStopFlag()
        self.assertFalse(pBDB.bibs.getArxivFieldsFlag)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class Test_Thread_importDailyArxiv(GUITestCase):
    """Test the functions in threadElements.Thread_importDailyArxiv"""

    def test_init(self):
        """test __init__"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_importDailyArxiv(ws, ["a", "b"], p)
        self.assertIsInstance(thr, PBThread)
        self.assertEqual(thr.parent(), p)
        self.assertEqual(thr.receiver, ws)
        self.assertEqual(thr.found, ["a", "b"])
        self.assertTrue(thr.runningImport)

    def test_run(self):
        """test run"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        self.assertTrue(ws.running)
        thr = Thread_importDailyArxiv(
            ws,
            {
                "12.345": {
                    "bibpars": {
                        "author": "me",
                        "title": "title",
                        "type": "",
                        "eprint": "12.345",
                        "replacement": False,
                        "cross": False,
                        "abstract": "some text",
                        "primaryclass": "astro-ph",
                    },
                    "exist": 1,
                }
            },
            p,
        )
        with patch(
            "physbiblio.database.Entries.loadAndInsert",
            return_value=True,
            autospec=True,
        ) as _lai, patch(
            "physbiblio.database.Entries.getByKey",
            return_value=[{"bibkey": "12.345"}],
            autospec=True,
        ) as _gbk, patch(
            "time.sleep"
        ) as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st, patch(
            "logging.Logger.info"
        ) as _in:
            thr.run()
            _lai.assert_called_once_with(pBDB.bibs, "12.345")
            _gbk.assert_called_once_with(pBDB.bibs, "12.345")
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
            _in.assert_has_calls(
                [
                    call("Entries successfully imported:\n['12.345']"),
                    call("Errors for entries:\n[]"),
                ]
            )

        q = Queue()
        ws = WriteStream(q)
        self.assertTrue(ws.running)
        thr = Thread_importDailyArxiv(
            ws,
            {
                "12.345": {
                    "bibpars": {
                        "author": "me1",
                        "title": "title1",
                        "type": "",
                        "eprint": "12.345",
                        "replacement": False,
                        "cross": False,
                        "abstract": "some text",
                        "primaryclass": "astro-ph",
                    },
                    "exist": 1,
                },
                "12.346": {
                    "bibpars": {
                        "author": "me2",
                        "title": "title2",
                        "type": "",
                        "eprint": "12.346",
                        "replacement": False,
                        "cross": True,
                        "abstract": "some other text",
                        "primaryclass": "astro-ph.CO",
                    },
                    "exist": 1,
                },
                "12.348": {
                    "bibpars": {
                        "author": "me4",
                        "title": "title4",
                        "type": "",
                        "eprint": "12.348",
                        "replacement": False,
                        "cross": False,
                        "abstract": "some more text",
                        "primaryclass": "hep-ph",
                    },
                    "exist": 1,
                },
                "12.349": {
                    "bibpars": {
                        "author": "me5",
                        "title": "title5",
                        "type": "",
                        "eprint": "12.349",
                        "replacement": False,
                        "cross": False,
                        "abstract": "some more text",
                        "primaryclass": "hep-ex",
                    },
                    "exist": 1,
                },
            },
            p,
        )
        with patch(
            "physbiblio.database.Entries.loadAndInsert",
            side_effect=[True, False, False, False],
            autospec=True,
        ) as _lai, patch(
            "physbiblio.database.Entries.prepareInsert",
            side_effect=["data1", "data2", "data3"],
            autospec=True,
        ) as _pi, patch(
            "physbiblio.database.Entries.updateInspireID",
            return_value="123",
            autospec=True,
        ) as _uid, patch(
            "physbiblio.database.Entries.searchOAIUpdates", autospec=True
        ) as _sou, patch(
            "physbiblio.database.Entries.insert",
            side_effect=[False, True, True],
            autospec=True,
        ) as _bi, patch(
            "physbiblio.database.Entries.getByKey",
            side_effect=[[{"bibkey": "12.345"}], [], [{"bibkey": "12.350"}]],
            autospec=True,
        ) as _gbk, patch(
            "physbiblio.database.Entries.getByBibkey",
            return_value=[{"bibkey": "12.346"}],
            autospec=True,
        ) as _gbb, patch(
            "logging.Logger.info"
        ) as _in, patch(
            "logging.Logger.warning"
        ) as _wa, patch(
            "time.sleep"
        ) as _sl, patch(
            "physbiblio.gui.commonClasses.WriteStream.start", autospec=True
        ) as _st:
            thr.run()
            self.assertFalse(ws.running)
            _st.assert_called_once_with(ws)
            _sl.assert_called_once_with(0.1)
            _lai.assert_has_calls(
                [
                    call(pBDB.bibs, "12.345"),
                    call(pBDB.bibs, "12.346"),
                    call(pBDB.bibs, "12.348"),
                    call(pBDB.bibs, "12.349"),
                ]
            )
            _gbk.assert_has_calls(
                [
                    call(pBDB.bibs, "12.345"),
                    call(pBDB.bibs, "12.348"),
                    call(pBDB.bibs, "12.349"),
                ]
            )
            _in.assert_has_calls(
                [
                    call("Element '12.348' successfully inserted.\n"),
                    call("Element '12.349' successfully inserted.\n"),
                    call(
                        "Entries successfully imported:\n"
                        + "['12.345', '12.348', '12.350']"
                    ),
                    call("Errors for entries:\n['12.346', '12.348']"),
                ]
            )
            _wa.assert_has_calls(
                [
                    call("Failed in inserting entry 12.346\n"),
                    call("Failed in completing info for entry 12.348\n"),
                ]
            )
            _pi.assert_has_calls(
                [
                    call(
                        pBDB.bibs,
                        '@Article{12.346,\n        author = "me2",\n'
                        + '         title = "{title2}",\n archiveprefix '
                        + '= "arXiv",\n  primaryclass = "astro-ph.CO",\n'
                        + '        eprint = "12.346",\n}\n\n',
                    ),
                    call(
                        pBDB.bibs,
                        '@Article{12.348,\n        author = "me4",\n'
                        + '         title = "{title4}",\n archiveprefix '
                        + '= "arXiv",\n  primaryclass = "hep-ph",\n'
                        + '        eprint = "12.348",\n}\n\n',
                    ),
                    call(
                        pBDB.bibs,
                        '@Article{12.349,\n        author = "me5",\n'
                        + '         title = "{title5}",\n archiveprefix '
                        + '= "arXiv",\n  primaryclass = "hep-ex",\n'
                        + '        eprint = "12.349",\n}\n\n',
                    ),
                ]
            )
            _bi.assert_has_calls([call(pBDB.bibs, "data1"), call(pBDB.bibs, "data2")])
            _uid.assert_has_calls(
                [call(pBDB.bibs, "12.348"), call(pBDB.bibs, "12.349")]
            )
            _sou.assert_has_calls(
                [
                    call(
                        pBDB.bibs,
                        0,
                        entries=[{"bibkey": "12.346"}],
                        force=True,
                        reloadAll=True,
                    ),
                    call(
                        pBDB.bibs,
                        0,
                        entries=[{"bibkey": "12.346"}],
                        force=True,
                        reloadAll=True,
                    ),
                ]
            )
            _gbb.assert_has_calls(
                [call(pBDB.bibs, "12.348"), call(pBDB.bibs, "12.349")]
            )

    def test_setStopFlag(self):
        """test setStopFlag"""
        p = QWidget()
        q = Queue()
        ws = WriteStream(q)
        thr = Thread_importDailyArxiv(ws, ["a", "b"], p)
        self.assertTrue(thr.runningImport)
        thr.setStopFlag()
        self.assertFalse(thr.runningImport)


if __name__ == "__main__":
    unittest.main()



"""Module that contains the pBErrorManager and pBLogger definitions.

This file is part of the physbiblio package.
"""
import sys
import traceback
import logging
import logging.handlers

try:
    from physbiblio.config import pbConfig
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class PBErrorManagerClass:
    """Class that manages the output of the errors and
    stores the messages into a log file
    """

    def __init__(self, loggerString="physbibliolog"):
        """Constructor for PBErrorManagerClass.

        Parameter:
            loggerString: the `Logger` identifier string to use
                (default="physbibliolog")
        """
        self.tempsh = []
        if pbConfig.params["loggingLevel"] == 0:
            self.loglevel = logging.ERROR
        elif pbConfig.params["loggingLevel"] == 1:
            self.loglevel = logging.WARNING
        elif pbConfig.params["loggingLevel"] == 2:
            self.loglevel = logging.INFO
        else:
            self.loglevel = logging.DEBUG
        # the main logger, will save to stdout and log file
        self.loggerString = loggerString
        self.logger = logging.getLogger(loggerString)
        self.logger.propagate = False
        self.logger.setLevel(min(logging.INFO, self.loglevel))

        try:
            fh = logging.handlers.RotatingFileHandler(
                pbConfig.overWritelogFileName, maxBytes=5.0 * 2 ** 20, backupCount=5
            )
        except AttributeError:
            fh = logging.handlers.RotatingFileHandler(
                pbConfig.params["logFileName"], maxBytes=5.0 * 2 ** 20, backupCount=5
            )
        fh.setLevel(self.loglevel)
        formatter = logging.Formatter(
            "%(asctime)s %(levelname)10s : [%(module)s.%(funcName)s] %(message)s"
        )
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        self.defaultStream = logging.StreamHandler()
        self.defaultStream.setLevel(min(logging.INFO, self.loglevel))
        formatter = logging.Formatter("[%(module)s.%(funcName)s] %(message)s")
        self.defaultStream.setFormatter(formatter)
        self.logger.addHandler(self.defaultStream)

    def tempHandler(
        self,
        stream=sys.stdout,
        level=logging.INFO,
        format="[%(module)s.%(funcName)s] %(message)s",
    ):
        """Set a temporary StreamHandler for the logger,
        given the stream, with default level logging.INFO.
        Useful when redirecting stdout

        Parameters:
            stream: the stream to be used (default: sys.stdout)
            level: the level to be used (default: logging.INFO)
            format: the format, using the logging syntax
                (default: '[%(module)s.%(funcName)s] %(message)s')
        """
        try:
            self.tempsh.append(logging.StreamHandler(stream))
            self.tempsh[-1].setLevel(level)
            formatter = logging.Formatter(format)
            self.tempsh[-1].setFormatter(formatter)
            self.logger.addHandler(self.tempsh[-1])
            return True
        except Exception:
            self.logger.exception("Failed while trying to add a new handler")
            return False

    def rmTempHandler(self):
        """Delete the last temporary handler that has been created"""
        self.logger.removeHandler(self.tempsh[-1])
        del self.tempsh[-1]

    def rmAllTempHandlers(self):
        """Delete the all the temporary handlers that have been created"""
        for temp in self.tempsh:
            self.logger.removeHandler(temp)
        self.tempsh = []

    def excepthook(self, cls, exception, trcbk):
        """Function that will replace `sys.excepthook` to log
        any error that occurs

        Parameters:
            cls, exception, trcbk as in `sys.excepthook`
        """
        self.logger.error("Unhandled exception", exc_info=(cls, exception, trcbk))


pBErrorManager = PBErrorManagerClass(pbConfig.loggerString)
pBLogger = pBErrorManager.logger
sys.excepthook = pBErrorManager.excepthook



"""Module that uses INSPIRE-HEP to get author
(number of papers, h index, citations) and paper statistics (citations).

Uses matplotlib to do plots.

This file is part of the physbiblio package.
"""
import time
import os
import traceback
import os.path as osp
import json
import requests
import dateutil
import datetime
import matplotlib

matplotlib.use("Qt5Agg")
os.environ["QT_API"] = "pyside2"
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.backends.backend_pdf import PdfPages

try:
    from physbiblio.errors import pBLogger
    from physbiblio.config import pbConfig
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class InspireStatsLoader:
    """Class that contains the methods
    to collect information from INSPIRE-HEP
    """

    urlBase = pbConfig.inspireSearchBase
    timeout = float(pbConfig.params["timeoutWebSearch"])
    authorStatsOpts = "&of=recjson&ot=recid,creation_date&so=a&rg=250"
    paperStatsOpts = "&of=recjson&ot=recid,creation_date&so=a&rg=250"
    skipPageOpt = "&jrec="
    maxPerPage = 250
    authorPlotInfo = {}
    paperPlotInfo = {}

    def __init__(self):
        """The class constructor,
        defines some constants and search options
        """
        self.urlBase = pbConfig.inspireSearchBase
        self.timeout = float(pbConfig.params["timeoutWebSearch"])
        self.authorPlotInfo = None
        self.paperPlotInfo = None
        self.allInfoA = {}
        self.authorPapersList = [[], []]
        self.allCitations = []
        self.runningAuthorStats = True
        self.runningPaperStats = True
        self.allInfoP = {}
        self.citingPapersList = [[], []]

    def changeBackend(self, wantBackend):
        """Changes the matplotlib backend currently in use.

        Parameters:
            wantBackend: a string that defines the wanted backend
        """
        if wantBackend != matplotlib.get_backend():
            matplotlib.use(wantBackend, warn=False, force=True)
            from matplotlib import pyplot as plt

            pBLogger.info("Changed backend to %s" % matplotlib.get_backend())

    def JsonFromUrl(self, url):
        """Function that downloads the url content
        and returns a Json object

        Parameters:
            url: string containing the url to be opened

        Output:
            the json object generated from the url content
        """

        def getSeries(url):
            response = requests.get(url, timeout=self.timeout)
            try:
                return json.loads(response.content.decode("utf-8"))
            except ValueError:
                pBLogger.warning("Empty response!")
                return []
            except:
                pBLogger.exception("Cannot read the page content!")
                return []

        ser = 0
        complete = []
        while True:
            temp = getSeries(
                url + self.skipPageOpt + "%d" % (ser * self.maxPerPage + 1)
            )
            if len(temp) < self.maxPerPage:
                return complete + temp
            else:
                complete += temp
                ser += 1

    def authorStats(self, authorName, plot=False, reset=True, pbMax=None, pbVal=None):
        """Function that gets the data and
        constructs the statistics for a given author.

        Parameters:
            authorName: the author name as identified into INSPIRE-HEP,
                or a list of author names
                (it calls itself recursively for all the list elements)
            plot (boolean, default False): True to call self.plotStats
            reset (boolean, default False): True to delete
                all previous existing data
                (used as False when processing a list of authors)
            pbMax (callable, optional): a function to set the maximum
                of a progress bar in the GUI, if possible
            pbVal (callable, optional): a function to set the value
                of a progress bar in the GUI, if possible

        Output:
            a dictionary containing all the statistic information.
            For a single author, the structure is the following:
            {
                "name": the author name,
                "aI": The complete information,
                    including the dictionaries with the single papers info
                    (see self.paperStats), the citations
                    and the corresponding dates,
                "paLi": a list of [id, date] of the papers
                    associated with the author,
                "allLi": the complete list of [date, total citations]
                    with all the citations to the papers,
                "meanLi": the complete list of
                    [date, total citations/number of papers]
                    computed at each point from "allLi" content,
                "h": the h-index,
                "figs" (only if `plot` is True): contains the figures.
                    See self.plotStats
            }
        """
        if reset:
            self.allInfoA = {}
            self.authorPapersList = [[], []]
            self.allCitations = []
        if isinstance(authorName, list):
            try:
                pbMax(len(authorName))
            except TypeError:
                pass
            for ia, a in enumerate(authorName):
                try:
                    pbVal(ia + 1)
                except TypeError:
                    pass
                self.authorStats(a, reset=False)
            self.authorPlotInfo["name"] = authorName
            return self.authorPlotInfo
        pBLogger.info("Stats for author '%s'" % authorName)
        url = (
            pbConfig.inspireSearchBase
            + "?p=author:"
            + authorName
            + self.authorStatsOpts
        )
        data = self.JsonFromUrl(url)
        recid_authorPapers = ["%d" % a["recid"] for a in data]
        tot = len(recid_authorPapers)
        pBLogger.info(
            "AuthorStats will process %d " % tot + "total papers to retrieve citations"
        )
        self.runningAuthorStats = True
        for i, p in enumerate(recid_authorPapers):
            if not self.runningAuthorStats:
                pBLogger.info(
                    "Received 'stop' signal. "
                    + "Interrupting download of information. "
                    + "Processing and exiting..."
                )
                break
            if tot > 20:
                time.sleep(1)
            if p in self.allInfoA.keys():
                continue
            self.allInfoA[p] = {}
            self.allInfoA[p]["date"] = dateutil.parser.parse(data[i]["creation_date"])
            self.authorPapersList[0].append(self.allInfoA[p]["date"])
            pBLogger.info(
                "%5d / %d (%5.2f%%) - looking for paper: '%s'"
                % (i + 1, tot, 100.0 * (i + 1) / tot, p)
            )
            paperInfo = self.paperStats(
                p, verbose=0, paperDate=self.allInfoA[p]["date"]
            )
            self.allInfoA[p]["infoDict"] = paperInfo["aI"]
            self.allInfoA[p]["citingPapersList"] = paperInfo["citList"]
            for c, v in self.allInfoA[p]["infoDict"].items():
                self.allCitations.append(v["date"])
            pBLogger.info("")
        self.authorPapersList[1] = []
        for i, p in enumerate(sorted(self.authorPapersList[0])):
            self.authorPapersList[0][i] = p
            self.authorPapersList[1].append(i + 1)
        pBLogger.info("Saving citation counts...")
        allCitList = [[], []]
        meanCitList = [[], []]
        currPaper = 0
        for i, d in enumerate(sorted(self.allCitations)):
            if (
                currPaper < len(self.authorPapersList[0]) - 1
                and d >= self.authorPapersList[0][currPaper + 1]
            ):
                currPaper += 1
            allCitList[0].append(d)
            allCitList[1].append(i + 1)
            meanCitList[0].append(d)
            meanCitList[1].append((i + 1.0) / self.authorPapersList[1][currPaper])
        hind = 0
        citations = [
            len(self.allInfoA[k]["citingPapersList"][0]) - 2
            for k in self.allInfoA.keys()
        ]
        for h in range(len(citations)):
            if len([a for a in citations if a >= h]) >= h:
                hind = h
        self.authorPlotInfo = {
            "name": authorName,
            "aI": self.allInfoA,
            "paLi": self.authorPapersList,
            "allLi": allCitList,
            "meanLi": meanCitList,
            "h": hind,
        }
        if plot:
            self.authorPlotInfo["figs"] = self.plotStats(author=True)
        pBLogger.info("Stats for author '%s' completed!" % authorName)
        return self.authorPlotInfo

    def paperStats(
        self,
        paperID,
        plot=False,
        verbose=1,
        paperDate=None,
        reset=True,
        pbMax=None,
        pbVal=None,
    ):
        """Function that gets the data and
        constructs the statistics for a given paper.

        Parameters:
            paperID (string): the INSPIRE-HEP id of the paper (a number)
            plot (boolean): whether or not the citations
                should be plotted (default False)
            verbose (int, default 1): increase the verbosity level
            paperDate (datetime, optional): the date of at which
                the paper was published
            reset (boolean, default False): True to delete
                all previous existing data
                (used as False when processing a list of IDs)
            pbMax (callable, optional): a function to set the maximum
                of a progress bar in the GUI, if possible
            pbVal (callable, optional): a function to set the value
                of a progress bar in the GUI, if possible

        Output:
            a dictionary containing all the desired information.
            The structure is the following:
            {
                "id": the paper ID,
                "aI": the list of creation date for all the papers,
                    in INSPIRE-HEP order,
                "citList": the ordered list of citing papers,
                "fig" (only if `plot` is True): contains the figure.
                    See self.plotStats
            }
        """
        if reset:
            self.allInfoP = {}
            self.citingPapersList = [[], []]
        if isinstance(paperID, list):
            self.runningPaperStats = True
            try:
                pbMax(len(paperID))
            except TypeError:
                pass
            for ia, a in enumerate(paperID):
                try:
                    pbVal(ia + 1)
                except TypeError:
                    pass
                if self.runningPaperStats:
                    self.paperStats(a, reset=False)
            self.paperPlotInfo["id"] = paperID
            return self.paperPlotInfo
        if verbose > 0:
            pBLogger.info("Stats for paper '%s'" % paperID)
        url = (
            pbConfig.inspireSearchBase
            + "?p=refersto:recid:"
            + paperID
            + self.paperStatsOpts
        )
        data = self.JsonFromUrl(url)
        recid_citingPapers = [a["recid"] for a in data]
        if paperDate is not None:
            self.citingPapersList[0].append(paperDate)
        for i, p in enumerate(recid_citingPapers):
            self.allInfoP[p] = {}
            self.allInfoP[p]["date"] = dateutil.parser.parse(data[i]["creation_date"])
            self.citingPapersList[0].append(self.allInfoP[p]["date"])
        for i, p in enumerate(sorted(self.citingPapersList[0])):
            self.citingPapersList[0][i] = p
            self.citingPapersList[1].append(i + 1)
        self.citingPapersList[0].append(
            datetime.datetime.fromordinal(datetime.date.today().toordinal())
        )
        try:
            self.citingPapersList[1].append(self.citingPapersList[1][-1])
        except IndexError:
            self.citingPapersList[1].append(0)
        self.paperPlotInfo = {
            "id": paperID,
            "aI": self.allInfoP,
            "citList": self.citingPapersList,
        }
        if plot:
            self.paperPlotInfo["fig"] = self.plotStats(paper=True)
        if verbose > 0:
            pBLogger.info("Done!")
        return self.paperPlotInfo

    def plotStats(
        self,
        paper=False,
        author=False,
        show=False,
        save=False,
        path=".",
        markPapers=False,
        pickVal=6,
    ):
        """Plot the collected information, using matplotlib.pyplot.

        Parameters:
            paper (boolean, default False): plot statistics
                for the last analyzed paper
            author (boolean, default False): plot statistics
                for the last analyzed author
            show (boolean, default False): True to show the plots
                in a separate window (with matplotlib.pyplot.show())
            save (boolean, default False): True to save the plots into files.
            path (string): where to save the plots
            markPapers (boolean, default False): True to draw
                a vertical lines at the dates
                corresponding to a paper appearing
            pickVal (float, default 6): the picker tolerance

        Output:
            False if paper==False and author==False,
            the matplotlib.pyplot figure containing
                the citation plot if paper==True,
            a list of matplotlib.pyplot figures containing
                the various plots if author==True
        """
        if paper and self.paperPlotInfo is not None:
            if len(self.paperPlotInfo["citList"][0]) > 0:
                pBLogger.info("Plotting for paper '%s'..." % self.paperPlotInfo["id"])
                fig, ax = plt.subplots()
                plt.plot(
                    self.paperPlotInfo["citList"][0],
                    self.paperPlotInfo["citList"][1],
                    picker=pickVal,
                )
                fig.autofmt_xdate()
                if save:
                    pdf = PdfPages(osp.join(path, self.paperPlotInfo["id"] + ".pdf"))
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                return fig
        elif author and self.authorPlotInfo is not None:
            pBLogger.info("Plotting for author '%s'..." % self.authorPlotInfo["name"])
            try:
                ymin = min(
                    int(self.authorPlotInfo["allLi"][0][0].strftime("%Y")) - 2,
                    int(self.authorPlotInfo["paLi"][0][0].strftime("%Y")) - 2,
                )
                ymax = max(
                    int(self.authorPlotInfo["allLi"][0][-1].strftime("%Y")) + 2,
                    int(self.authorPlotInfo["paLi"][0][-1].strftime("%Y")) + 2,
                )
            except:
                try:
                    ymin = int(self.authorPlotInfo["paLi"][0][0].strftime("%Y")) - 2
                    ymax = int(self.authorPlotInfo["paLi"][0][-1].strftime("%Y")) + 2
                except:
                    pBLogger.warning("No publications for this author?")
                    return False
            figs = []
            if len(self.authorPlotInfo["paLi"][0]) > 0:
                fig, ax = plt.subplots()
                plt.title("Paper number")
                plt.plot(
                    self.authorPlotInfo["paLi"][0],
                    self.authorPlotInfo["paLi"][1],
                    picker=pickVal,
                )
                fig.autofmt_xdate()
                if save:
                    pdf = PdfPages(
                        osp.join(path, self.authorPlotInfo["name"] + "_papers.pdf")
                    )
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                figs.append(fig)

            if len(self.authorPlotInfo["paLi"][0]) > 0:
                fig, ax = plt.subplots()
                plt.title("Papers per year")
                ax.hist(
                    [int(q.strftime("%Y")) for q in self.authorPlotInfo["paLi"][0]],
                    bins=range(ymin, ymax),
                    picker=True,
                )
                ax.get_xaxis().get_major_formatter().set_useOffset(False)
                plt.xlim([ymin, ymax])
                if save:
                    pdf = PdfPages(
                        osp.join(path, self.authorPlotInfo["name"] + "_yearPapers.pdf")
                    )
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                figs.append(fig)

            if len(self.authorPlotInfo["allLi"][0]) > 0:
                fig, ax = plt.subplots()
                plt.title("Total citations")
                plt.plot(
                    self.authorPlotInfo["allLi"][0],
                    self.authorPlotInfo["allLi"][1],
                    picker=pickVal,
                )
                fig.autofmt_xdate()
                if save:
                    pdf = PdfPages(
                        osp.join(path, self.authorPlotInfo["name"] + "_allCit.pdf")
                    )
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                figs.append(fig)

            if len(self.authorPlotInfo["allLi"][0]) > 0:
                fig, ax = plt.subplots()
                plt.title("Citations per year")
                ax.hist(
                    [int(q.strftime("%Y")) for q in self.authorPlotInfo["allLi"][0]],
                    bins=range(ymin, ymax),
                    picker=True,
                )
                ax.get_xaxis().get_major_formatter().set_useOffset(False)
                plt.xlim([ymin, ymax])
                if save:
                    pdf = PdfPages(
                        osp.join(path, self.authorPlotInfo["name"] + "_yearCit.pdf")
                    )
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                figs.append(fig)

            if len(self.authorPlotInfo["meanLi"][0]) > 0:
                fig, ax = plt.subplots()
                plt.title("Mean citations")
                plt.plot(
                    self.authorPlotInfo["meanLi"][0],
                    self.authorPlotInfo["meanLi"][1],
                    picker=pickVal,
                )
                fig.autofmt_xdate()
                if markPapers:
                    for q in self.authorPlotInfo["paLi"][0]:
                        plt.axvline(
                            datetime.datetime(
                                int(q.strftime("%Y")),
                                int(q.strftime("%m")),
                                int(q.strftime("%d")),
                            ),
                            color="k",
                            ls="--",
                        )
                if save:
                    pdf = PdfPages(
                        osp.join(path, self.authorPlotInfo["name"] + "_meanCit.pdf")
                    )
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                figs.append(fig)

            if len(self.authorPlotInfo["aI"].keys()) > 0:
                fig, ax = plt.subplots()
                plt.title("Citations for each paper")
                for i, p in enumerate(self.authorPlotInfo["aI"].keys()):
                    try:
                        plt.plot(
                            self.authorPlotInfo["aI"][p]["citingPapersList"][0],
                            self.authorPlotInfo["aI"][p]["citingPapersList"][1],
                        )
                    except:
                        pBLogger.exception("Something went wrong while plotting...")
                fig.autofmt_xdate()
                if save:
                    pdf = PdfPages(
                        osp.join(path, self.authorPlotInfo["name"] + "_paperCit.pdf")
                    )
                    pdf.savefig()
                    pdf.close()
                if show:
                    plt.show()
                plt.close()
                figs.append(fig)
            return figs
        else:
            pBLogger.info("Nothing to plot...")
            return False


pBStats = InspireStatsLoader()



"""Module that creates the base class WebInterf,
which will be used by other modules in this package.

Uses urllib to download url content.

This file is part of the physbiblio package.
"""
import sys
import os
import socket
import pkgutil
import traceback
import ssl

if sys.version_info[0] < 3:
    from urllib2 import Request, urlopen, URLError, HTTPError
else:
    from urllib.request import Request, urlopen, URLError, HTTPError

try:
    from physbiblio.errors import pBLogger
    import physbiblio.webimport as wi
    from physbiblio.config import pbConfig
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise

# scan package content to load list of available modules, to be imported later
pkgpath = os.path.dirname(wi.__file__)
webInterfaces = [name for _, name, _ in pkgutil.iter_modules([pkgpath])]


class WebInterf:
    """This is the main class for the web search methods.

    It contains a constructor, a function to create an appropriate url
    and to retrieve text from the url,
    a function to load other webinterfaces
    """

    url = None
    urlArgs = None
    urlTimeout = 1000.0
    interfaces = []
    webSearch = {}
    loaded = False

    def __init__(self):
        """Initializes the class variables."""
        self.url = None
        self.urlArgs = None
        self.urlTimeout = float(pbConfig.params["timeoutWebSearch"])
        # save the names of the available web search interfaces
        self.interfaces = [
            a for a in webInterfaces if a != "webInterf" and a != "tests"
        ]
        self.webSearch = {}
        self.loaded = False

    def createUrl(self):
        """Joins the arguments of the GET query to get the full url.

        Uses the self.urlArgs dictionary to generate the list
        of HTTP GET parameters.
        """
        return (
            self.url
            + "?"
            + "&".join(["%s=%s" % (a, b) for a, b in self.urlArgs.items()])
        )

    def textFromUrl(self, url, headers=None):
        """Use urllib to get the html content of the given url.

        Parameters:
            url: the url to be opened
            headers (default None): the additional headers
                to be passed to urllib.Request

        Output:
            text: the content of the url
        """
        try:
            if headers is not None:
                req = Request(url, headers=headers)
            else:
                req = Request(url)
            response = urlopen(req, timeout=self.urlTimeout)
            data = response.read()
        except URLError:
            pBLogger.warning("[%s] -> Error in retrieving data from url" % self.name)
            return ""
        except HTTPError:
            pBLogger.warning("[%s] -> %s not found" % url)
            return ""
        except (ssl.SSLError, socket.timeout):
            pBLogger.warning("[%s] -> Timed out" % self.name)
            return ""
        try:
            text = data.decode("utf-8")
        except Exception:
            pBLogger.warning(
                "[%s] -> Bad codification, utf-8 decode failed" % self.name
            )
            return ""
        return text

    def retrieveUrlFirst(self, search):
        """Retrieves the first bibtexs that the search gives,
        using the subclass specific instructions.

        Parameter:
            search: the string to be searched

        Output:
            returns None in the default implementation
                (must be subclassed)
        """
        return None

    def retrieveUrlAll(self, search):
        """Retrieves all the bibtexs that the search gives,
        using the subclass specific instructions

        Parameter:
            search: the string to be searched

        Output:
            returns None in the default implementation
                (must be subclassed)
        """
        return None

    def loadInterfaces(self):
        """Load the subclasses that will interface
        with the main websites to search bibtex info
        and saves them into a dictionary (`self.webSearch`).

        The subclasses are read scanning the package directory.
        """
        if self.loaded:
            return
        for method in self.interfaces:
            try:
                _temp = __import__(
                    "physbiblio.webimport." + method, globals(), locals(), ["WebSearch"]
                )
                self.webSearch[method] = getattr(_temp, "WebSearch")()
            except Exception:
                pBLogger.exception("Error importing physbiblio.webimport.%s" % method)
        self.loaded = True

    def retrieveUrlFirstFrom(self, search, method):
        """Calls the function retrieveUrlFirst
        given the subclass method.

        Parameters:
            search: the search string
            method: the key of the method in `self.webSearch`

        Output:
            None in the default implementation (must be subclassed)
        """
        try:
            return getattr(self.webSearch[method], retrieveUrlFirst)(search)
        except KeyError:
            pBLogger.warning("The method '%s' is not available!" % method)
            return ""

    def retrieveUrlAllFrom(self, search, method):
        """Calls the function retrieveUrlAll given the subclass method.

        Parameters:
            search: the search string
            method: the key of the method in `self.webSearch`

        Output:
            None in the default implementation (must be subclassed)
        """
        try:
            return getattr(self.webSearch[method], retrieveUrlAll)(search)
        except KeyError:
            pBLogger.warning("The method '%s' is not available!" % method)
            return ""


physBiblioWeb = WebInterf()
physBiblioWeb.loadInterfaces()



#!/usr/bin/env python
"""Test file for the physbiblio.argParser module.

This file is part of the physbiblio package.
"""
import sys
import traceback

if sys.version_info[0] < 3:
    import unittest2 as unittest
    from mock import MagicMock, patch
    from StringIO import StringIO
else:
    import unittest
    from unittest.mock import MagicMock, patch
    from io import StringIO

from PySide2.QtWidgets import QApplication

try:
    from physbiblio import __version__, __version_date__
    from physbiblio.setuptests import *
    from physbiblio.gui.setuptests import *
    from physbiblio.argParser import *
    from physbiblio.config import pbConfig
    from physbiblio.database import pBDB
    from physbiblio.export import pBExport
    from physbiblio.gui.mainWindow import MainWindow
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


class TestParser(unittest.TestCase):
    """Tests for main parser and stuff"""

    @patch("sys.stdout", new_callable=StringIO)
    def assert_in_stdout_sysexit(self, function, expected_output, mock_stdout):
        """Catch and if test stdout of the function contains a string"""
        with self.assertRaises(SystemExit):
            function()
        string = mock_stdout.getvalue()
        for text in expected_output:
            self.assertIn(text, string)

    @patch("sys.stderr", new_callable=StringIO)
    def assert_in_stderr_sysexit(self, function, expected_output, mock_stderr):
        """Catch and if test stderr of the function contains a string"""
        with self.assertRaises(SystemExit):
            function()
        string = mock_stderr.getvalue()
        for text in expected_output:
            self.assertIn(text, string)

    def test_global(self):
        """Test that the options are recognised correctly"""
        parser = setParser()
        for opt in ["-v", "--version"]:
            with self.assertRaises(SystemExit):
                parser.parse_args([opt])
            if sys.version_info[0] < 3:
                self.assert_in_stderr_sysexit(
                    lambda: parser.parse_args([opt]),
                    ["PhysBiblio %s (%s)" % (__version__, __version_date__)],
                )
            else:
                self.assert_in_stdout_sysexit(
                    lambda: parser.parse_args([opt]),
                    ["PhysBiblio %s (%s)" % (__version__, __version_date__)],
                )
        for opt in ["-h", "--help"]:
            with self.assertRaises(SystemExit):
                parser.parse_args([opt])
            self.assert_in_stdout_sysexit(
                lambda: parser.parse_args([opt]),
                [
                    "usage: PhysBiblio.exe [-h] [-p ",
                    "{clean,cli,daily,dates,export,test,tex,update,weekly,gui}",
                ],
            )
        for opt in ["-p", "--profile"]:
            profile = list(pbConfig.profiles.keys())[0]
            with patch("physbiblio.cli.cli", autospec=True) as mock, patch(
                "logging.Logger.info"
            ) as _i:
                parser.parse_args([opt, "%s" % profile, "cli"])
                self.assertIn(
                    "Starting with profile '%s', database" % profile, _i.call_args[0][0]
                )
                args = parser.parse_args([opt, "%s" % profile, "cli"])
                args.func(args)
                mock.assert_called_once_with()

    def test_subcommands_ok(self):
        """Test that the options are recognised correctly"""
        parser = setParser()
        tests = [["gui", call_gui], ["test", call_tests]]
        for sub, func in tests:
            args = parser.parse_args([sub])
            self.assertIs(args.func, func)
        tests = [
            [
                "clean",
                "physbiblio.database.Entries.cleanBibtexs",
                [],
                ([pBDB.bibs], {"startFrom": 0}),
            ],
            [
                "clean",
                "physbiblio.database.Entries.cleanBibtexs",
                ["-s", "100"],
                ([pBDB.bibs], {"startFrom": 100}),
            ],
            [
                "clean",
                "physbiblio.database.Entries.cleanBibtexs",
                ["--startFrom", "1000"],
                ([pBDB.bibs], {"startFrom": 1000}),
            ],
            ["cli", "physbiblio.cli.cli", [], ([], {})],
            [
                "daily",
                "physbiblio.database.Entries.getDailyInfoFromOAI",
                [],
                (
                    [
                        pBDB.bibs,
                        (datetime.date.today() - datetime.timedelta(1)).strftime(
                            "%Y-%m-%d"
                        ),
                        datetime.date.today().strftime("%Y-%m-%d"),
                    ],
                    {},
                ),
            ],
            [
                "dates",
                "physbiblio.database.Entries.getDailyInfoFromOAI",
                ["2018-04-01", "2018-05-01"],
                ([pBDB.bibs, "2018-04-01", "2018-05-01"], {}),
            ],
            [
                "export",
                "physbiblio.export.PBExport.exportAll",
                ["testname"],
                ([pBExport, "testname"], {}),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "-o"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": True,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "--overwrite"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": True,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "--ov"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": True,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "-s"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": False,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "--save"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": False,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "--sa"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": False,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "f3", "--sa"],
                (
                    [pBExport, ["f2", "f3"], "f1"],
                    {
                        "autosave": False,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "-u"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": True,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "--updateExisting"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": False,
                        "removeUnused": False,
                        "updateExisting": True,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "-r"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": False,
                        "removeUnused": True,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "tex",
                "physbiblio.export.PBExport.exportForTexFile",
                ["f1", "f2", "--removeUnused"],
                (
                    [pBExport, ["f2"], "f1"],
                    {
                        "autosave": True,
                        "overwrite": False,
                        "removeUnused": True,
                        "updateExisting": False,
                    },
                ),
            ],
            [
                "update",
                "physbiblio.database.Entries.searchOAIUpdates",
                [],
                ([pBDB.bibs], {"startFrom": 0, "force": False}),
            ],
            [
                "update",
                "physbiblio.database.Entries.searchOAIUpdates",
                ["-f"],
                ([pBDB.bibs], {"startFrom": 0, "force": True}),
            ],
            [
                "update",
                "physbiblio.database.Entries.searchOAIUpdates",
                ["--force"],
                ([pBDB.bibs], {"startFrom": 0, "force": True}),
            ],
            [
                "update",
                "physbiblio.database.Entries.searchOAIUpdates",
                ["-s", "100"],
                ([pBDB.bibs], {"startFrom": 100, "force": False}),
            ],
            [
                "update",
                "physbiblio.database.Entries.searchOAIUpdates",
                ["--startFrom", "100"],
                ([pBDB.bibs], {"startFrom": 100, "force": False}),
            ],
            [
                "weekly",
                "physbiblio.database.Entries.getDailyInfoFromOAI",
                [],
                (
                    [
                        pBDB.bibs,
                        (datetime.date.today() - datetime.timedelta(7)).strftime(
                            "%Y-%m-%d"
                        ),
                        datetime.date.today().strftime("%Y-%m-%d"),
                    ],
                    {},
                ),
            ],
        ]
        with patch(
            "physbiblio.database.PhysBiblioDB.commit", autospec=True
        ) as _comm_mock:
            for sub, func, options, expected in tests:
                with patch(func, autospec=True) as mock:
                    args = parser.parse_args([sub] + options)
                    args.func(args)
                    mock.assert_called_once_with(*expected[0], **expected[1])

    def test_subcommands_fail(self):
        """Test that the options are recognised correctly."""
        parser = setParser()
        tests = [
            ["clean", ["-f"]],
            ["clean", ["-s"]],
            ["clean", ["-s", "abc"]],
            ["clean", ["--startfrom", "1000"]],
            ["cli", ["-t"]],
            ["daily", ["date"]],
            ["dates", ["date1"]],
            ["export", ["testname1", "testname2"]],
            ["gui", ["-p"]],
            ["test", ["-f"]],
            ["tex", ["f1"]],
            ["tex", ["f1", "f2", "--overWrite"]],
            ["tex", ["f1", "f2", "-save"]],
            ["update", ["-o"]],
            ["update", ["-s", "abc"]],
            ["update", ["--startFrom"]],
            ["weekly", ["any"]],
        ]
        with patch(
            "physbiblio.database.PhysBiblioDB.commit", autospec=True
        ) as _comm_mock:
            for sub, options in tests:
                with self.assertRaises(SystemExit):
                    args = parser.parse_args([sub] + options)

    def test_subcommand_tests(self):
        """Test that the options are recognised correctly"""
        from physbiblio.setuptests import skipTestsSettings

        if sys.version_info[0] < 3:
            patchString = "unittest2.runner.TextTestRunner.run"
        else:
            patchString = "unittest.runner.TextTestRunner.run"
        oldSettings = skipTestsSettings.copy()
        parser = setParser()
        tests = [
            [["test"], False, False, False, False, False],
            [["test", "-d"], True, False, False, False, False],
            [["test", "--database"], True, False, False, False, False],
            [["test", "-g"], False, True, False, False, False],
            [["test", "--gui"], False, True, False, False, False],
            [["test", "-l"], False, False, True, False, False],
            [["test", "--long"], False, False, True, False, False],
            [["test", "-o"], False, False, False, True, True],
            [["test", "--online"], False, False, False, True, True],
            [["test", "--online_oai"], False, False, False, True, False],
            [["test", "-d", "--online"], True, False, False, True, True],
        ]
        for options, _db, _gui, _lon, _oai, _onl in tests:
            skipTestsSettings.default()
            with patch(patchString, autospec=True) as _run:
                args = parser.parse_args(options)
                call_tests(args)
                self.assertEqual(skipTestsSettings.db, _db)
                self.assertEqual(skipTestsSettings.gui, _gui)
                self.assertEqual(skipTestsSettings.long, _lon)
                self.assertEqual(skipTestsSettings.oai, _oai)
                self.assertEqual(skipTestsSettings.online, _onl)
                self.assertEqual(_run.call_count, 1)
        with self.assertRaises(SystemExit):
            args = parser.parse_args(["test", "-a"])
        skipTestsSettings = oldSettings

    def test_call_gui(self):
        """test the call_gui function"""
        mw = MainWindow()
        mw.show = MagicMock()
        mw.raise_ = MagicMock()
        mw.recentChanges = MagicMock()
        mw.errormessage = MagicMock()
        with patch(
            "physbiblio.argParser.QApplication", return_value=globalQApp
        ) as _qa, patch(
            "physbiblio.gui.mainWindow.MainWindow", return_value=mw
        ) as _mw, patch(
            "physbiblio.config.ConfigurationDB.update"
        ) as _u, patch(
            "physbiblio.config.GlobalDB.commit"
        ) as _c, patch(
            "sys.exit"
        ) as _e:
            with patch.dict(
                pbConfig.params, {"openSinceLastUpdate": __version__}, clear=False
            ):
                call_gui([])
            _qa.assert_called_once_with(sys.argv)
            _mw.assert_called_once_with()
            mw.show.assert_called_once_with()
            mw.raise_.assert_called_once_with()
            _e.assert_called_once_with(0)
            self.assertEqual(mw.recentChanges.call_count, 0)
            with patch.dict(
                pbConfig.params,
                {"openSinceLastUpdate": __version__ + "abc"},
                clear=False,
            ):
                call_gui([])
            mw.recentChanges.assert_called_once_with()
            _u.assert_called_once_with("openSinceLastUpdate", __version__)
            _c.assert_called_once_with()

    @classmethod
    def tearDownModule(self):
        sys.excepthook = sys.__excepthook__


if __name__ == "__main__":
    unittest.main()



"""Module that deals with importing info from the DOI.org API.

This file is part of the physbiblio package.
"""
import traceback

try:
    from physbiblio.errors import pBLogger
    from physbiblio.webimport.webInterf import WebInterf
    from physbiblio.parseAccents import parse_accents_str
    from physbiblio.config import pbConfig
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class WebSearch(WebInterf):
    """Subclass of WebInterf that can connect
    to doi.org to perform searches
    """

    name = "doi"
    description = "Doi fetcher"
    url = pbConfig.doiUrl
    headers = {"accept": "application/x-bibtex"}

    def __init__(self):
        """Initializes the class variables using the WebInterf constructor.

        Define additional specific parameters for the DOI.org API.
        """
        WebInterf.__init__(self)
        self.name = "doi"
        self.description = "Doi fetcher"
        self.url = pbConfig.doiUrl
        self.headers = {"accept": "application/x-bibtex"}

    def createUrl(self, doi):
        """Joins the base url and the search string to get the full url.

        (DOI.org url Behaves differently than other APIs
        in the modules of this subpackage)
        """
        return self.url + doi

    def retrieveUrlFirst(self, string):
        """Retrieves the first (only) result from the content
        of the given web page.

        Parameters:
            string: the search string (the DOI)

        Output:
            returns the bibtex string
        """
        url = self.createUrl(string)
        pBLogger.info("Search '%s' -> %s" % (string, url))
        text = self.textFromUrl(url, self.headers)
        try:
            return parse_accents_str(text[:])
        except Exception:
            pBLogger.exception("Impossible to get results")
            return ""

    def retrieveUrlAll(self, string):
        """Alias for retrieveUrlFirst
        (no more than one object should match a doi)
        """
        return self.retrieveUrlFirst(string)



"""Use INSPIRE-HEP OAI API to collect information on single papers
(given the identifier) or to harvest updates in a given time period.

This file is part of the physbiblio package.
"""
import sys
import time
import codecs
import re

if sys.version_info[0] < 3:
    # needed to set utf-8 as encoding
    reload(sys)
    sys.setdefaultencoding("utf-8")
    from httplib import IncompleteRead
    from urllib2 import URLError
else:
    from http.client import IncompleteRead
    from urllib.request import URLError

import datetime
import traceback

import bibtexparser
from oaipmh.client import Client
from oaipmh.error import ErrorBase
from oaipmh.metadata import MetadataRegistry

if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO
from lxml.etree import tostring
from pymarc import marcxml, MARCWriter, field
from oaipmh import metadata
from socket import error as SocketError

try:
    from physbiblio.errors import pBLogger
    from physbiblio.config import pbConfig
    from physbiblio.webimport.webInterf import WebInterf
    from physbiblio.webimport.arxiv import getYear
    from physbiblio.parseAccents import parse_accents_str
    from bibtexparser.bibdatabase import BibDatabase
    from physbiblio.bibtexWriter import pbWriter
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


def safe_list_get(l, idx, default=""):
    """Safely get an element from a list.
    No error if it doesn't exist...

    Parameters:
        l: the list from which to get the element
        idx: the index
        default: the default return value if `l[idx]` does not exist
    """
    if l is not None:
        try:
            return l[idx]
        except IndexError:
            return default
    else:
        return default


def get_journal_ref_xml(marcxml):
    """Read the marcxml record and write the info on the publication

    Parameter:
        marcxml: the marcxml record to read
    """
    p = []
    y = []
    v = []
    c = []
    m = []
    x = []
    t = []
    w = []
    if marcxml["773"] is not None:
        for q in marcxml.get_fields("773"):
            # journal name (even if submitted to only)
            p.append(parse_accents_str(q["p"]))
            v.append(parse_accents_str(q["v"]))  # volume
            y.append(parse_accents_str(q["y"]))  # year
            c.append(parse_accents_str(q["c"]))  # pages

            # Erratum, Addendum, Publisher note
            m.append(parse_accents_str(q["m"]))

            # freetext journal/book info
            x.append(parse_accents_str(q["x"]))
            # for conf papers: presented at etc, freetext or KB?
            t.append(parse_accents_str(q["t"]))
            # for conf papers: conference code
            w.append(parse_accents_str(q["w"]))
    return p, v, y, c, m, x, t, w


class MARCXMLReader(object):
    """Returns the PyMARC record from the OAI structure for MARC XML"""

    def __call__(self, element):
        """Call the xml parser and return the records

        Parameter:
            element: the xml text to read
        """
        handler = marcxml.XmlHandler()
        if sys.version_info[0] < 3:
            marcxml.parse_xml(StringIO(tostring(element[0], encoding="UTF-8")), handler)
        else:
            marcxml.parse_xml(StringIO(tostring(element[0], encoding=str)), handler)
        return handler.records[0]


marcxml_reader = MARCXMLReader()

registry = metadata.MetadataRegistry()
registry.registerReader("marcxml", marcxml_reader)


class WebSearch(WebInterf):
    """Subclass of WebInterf that can connect to INSPIRE-HEP
    to perform searches using the OAI API
    """

    name = "inspireoai"
    description = "INSPIRE OAI interface"
    url = "https://inspirehep.net/oai2d"
    correspondences = [
        ["id", "inspire"],
        ["year", "year"],
        ["arxiv", "arxiv"],
        # ["oldkeys", "old_keys"],
        ["firstdate", "firstdate"],
        ["pubdate", "pubdate"],
        ["doi", "doi"],
        ["ads", "ads"],
        ["isbn", "isbn"],
        ["bibtex", "bibtex"],
        ["link", "link"],
    ]
    bibtexFields = [
        "author",
        "title",
        "journal",
        "volume",
        "year",
        "pages",
        "arxiv",
        "primaryclass",
        "archiveprefix",
        "eprint",
        "doi",
        "isbn",
        "school",
        "reportnumber",
        "booktitle",
        "collaboration",
    ]

    def __init__(self):
        """Initializes the class variables using
        the WebInterf constructor.

        Define additional specific parameters
        for the INSPIRE-HEP OAI API.
        """
        WebInterf.__init__(self)
        self.name = "inspireoai"
        self.description = "INSPIRE OAI interface"
        self.url = "https://inspirehep.net/oai2d"
        self.oai = Client(self.url, registry)

    def retrieveUrlFirst(self, string):
        """The OAI interface is not for string searches:
        use the retrieveOAIData function if you have the INSPIRE ID
        of the desired record
        """
        pBLogger.warning("Inspireoai cannot search strings in the DB")
        return ""

    def retrieveUrlAll(self, string):
        """The OAI interface is not for string searches:
        use the retrieveOAIData function if you have the INSPIRE ID
        of the desired record
        """
        pBLogger.warning("Inspireoai cannot search strings in the DB")
        return ""

    def readRecord(self, record, readConferenceTitle=False):
        """Read the content of a marcxml record
        to return a bibtex string

        Parameters:
            record: the marcxml record to read
            readConferenceTitle (default False): if True, look for
                the proceedings info to get the title of the conference

        Output:
            a dictionary with the obtained fields
        """
        tmpDict = {}
        record.to_unicode = True
        record.force_utf8 = True
        arxiv = ""
        tmpOld = []
        try:
            tmpDict["doi"] = None
            for q in record.get_fields("024"):
                if q["2"] == "DOI":
                    tmpDict["doi"] = q["a"]
        except Exception as e:
            pBLogger.warning("Error in readRecord!", exc_info=True)
        try:
            tmpDict["arxiv"] = None
            tmpDict["bibkey"] = None
            tmpDict["ads"] = None
            for q in record.get_fields("035"):
                if q["9"] == "arXiv":
                    tmp = q["a"]
                    if tmp is not None:
                        arxiv = tmp.replace("oai:arXiv.org:", "")
                    else:
                        arxiv = ""
                    tmpDict["arxiv"] = arxiv
                if q["9"] == "SPIRESTeX" or q["9"] == "INSPIRETeX":
                    if q["z"]:
                        tmpDict["bibkey"] = q["z"]
                    elif q["a"]:
                        tmpOld.append(q["a"])
                if q["9"] == "ADS":
                    if q["a"] is not None:
                        tmpDict["ads"] = q["a"]
        except (IndexError, TypeError) as e:
            pBLogger.warning("Error in readRecord!", exc_info=True)
        if tmpDict["bibkey"] is None and len(tmpOld) > 0:
            tmpDict["bibkey"] = tmpOld[0]
            tmpOld = []
        try:
            j, v, y, p, m, x, t, w = get_journal_ref_xml(record)
            tmpDict["journal"] = j[0]
            tmpDict["volume"] = v[0]
            tmpDict["year"] = y[0]
            tmpDict["pages"] = p[0]
            if w[0] is not None:
                conferenceCode = w[0]
            else:
                conferenceCode = None
        except IndexError:
            tmpDict["journal"] = None
            tmpDict["volume"] = None
            tmpDict["year"] = None
            tmpDict["pages"] = None
            conferenceCode = None
        try:
            firstdate = record["269"]
            if firstdate is not None:
                firstdate = firstdate["c"]
            else:
                firstdate = record["961"]
                if firstdate is not None:
                    firstdate = firstdate["x"]
            tmpDict["firstdate"] = firstdate
        except TypeError:
            tmpDict["firstdate"] = None
        try:
            tmpDict["pubdate"] = record["260"]["c"]
        except TypeError:
            tmpDict["pubdate"] = None
        try:
            tmpDict["author"] = record["100"]["a"]
        except TypeError:
            tmpDict["author"] = ""
        try:
            addAuthors = 0
            if len(record.get_fields("700")) > pbConfig.params["maxAuthorSave"] - 1:
                tmpDict["author"] += " and others"
            else:
                for r in record.get_fields("700"):
                    addAuthors += 1
                    if addAuthors > pbConfig.params["maxAuthorSave"]:
                        tmpDict["author"] += " and others"
                        break
                    tmpDict["author"] += " and %s" % r["a"]
        except:
            pass
        try:
            tmpDict["collaboration"] = record["710"]["g"]
        except TypeError:
            tmpDict["collaboration"] = None
        tmpDict["primaryclass"] = None
        tmpDict["archiveprefix"] = None
        tmpDict["eprint"] = None
        tmpDict["reportnumber"] = None
        try:
            for q in record.get_fields("037"):
                if "arXiv" in q["a"]:
                    tmpDict["primaryclass"] = q["c"]
                    tmpDict["archiveprefix"] = q["9"]
                    tmpDict["eprint"] = q["a"].lower().replace("arxiv:", "")
                else:
                    tmpDict["reportnumber"] = q["a"]
        except:
            pass
        if tmpDict["arxiv"] != tmpDict["eprint"] and tmpDict["eprint"] is not None:
            tmpDict["arxiv"] = tmpDict["eprint"]
        if tmpDict["eprint"] is not None and tmpDict["year"] is None:
            tmpDict["year"] = getYear(tmpDict["eprint"])
        try:
            tmpDict["title"] = record["245"]["a"]
        except TypeError:
            tmpDict["title"] = None
        try:
            tmpDict["isbn"] = record["020"]["a"]
        except TypeError:
            tmpDict["isbn"] = None
        if conferenceCode is not None and readConferenceTitle:
            url = (
                pbConfig.inspireSearchBase
                + "?p=773__w%%3A%s+or+773__w%%3A%s+and+980__a%%3A"
                % (conferenceCode, conferenceCode.replace("-", "%2F"))
                + "Proceedings&of=xe"
            )
            text = parse_accents_str(self.textFromUrl(url))
            title = re.compile("<title>(.*)</title>", re.MULTILINE)
            try:
                tmpDict["booktitle"] = [m.group(1) for m in title.finditer(text)][0]
            except IndexError:
                tmpDict["booktitle"] = None
        if tmpDict["isbn"] is not None:
            tmpDict["ENTRYTYPE"] = "book"
        else:
            try:
                collections = [r["a"].lower() for r in record.get_fields("980")]
            except:
                collections = []
            if "conferencepaper" in collections or conferenceCode is not None:
                tmpDict["ENTRYTYPE"] = "inproceedings"
            elif "thesis" in collections:
                tmpDict["ENTRYTYPE"] = "phdthesis"
                try:
                    tmpDict["school"] = record["502"]["c"]
                    tmpDict["year"] = record["502"]["d"]
                except:
                    pass
            else:
                tmpDict["ENTRYTYPE"] = "article"
        tmpDict["oldkeys"] = ",".join(tmpOld)
        for k in tmpDict.keys():
            try:
                tmpDict[k] = parse_accents_str(tmpDict[k])
            except:
                pass
        tmpDict["link"] = None
        try:
            if tmpDict["arxiv"] is not None and tmpDict["arxiv"] != "":
                tmpDict["link"] = pbConfig.arxivUrl + "/abs/" + tmpDict["arxiv"]
        except KeyError:
            pass
        try:
            if tmpDict["doi"] is not None and tmpDict["doi"] != "":
                tmpDict["link"] = pbConfig.doiUrl + tmpDict["doi"]
        except KeyError:
            pass
        bibtexDict = {"ENTRYTYPE": tmpDict["ENTRYTYPE"], "ID": tmpDict["bibkey"]}
        for k in self.bibtexFields:
            if k in tmpDict.keys() and tmpDict[k] is not None and tmpDict[k] is not "":
                bibtexDict[k] = tmpDict[k]
        try:
            if (
                bibtexDict["arxiv"] == bibtexDict["eprint"]
                and bibtexDict["eprint"] is not None
            ):
                del bibtexDict["arxiv"]
        except:
            pass
        db = bibtexparser.bibdatabase.BibDatabase()
        db.entries = [bibtexDict]
        tmpDict["bibtex"] = pbWriter.write(db)
        return tmpDict

    def retrieveOAIData(
        self, inspireID, bibtex=None, verbose=0, readConferenceTitle=False
    ):
        """Get the marcxml entry for a given record

        Parameters:
            inspireID: the INSPIRE-HEP identifier (a number)
                of the desired entry
            bibtex (default None): whether the bibtex should be
                included in the output dictionary
            verbose (default 0): increase the output level
            readConferenceTitle (boolean, default False):
                try to read the conference title if dealing
                with a proceeding

        Output:
            the dictionary containing the bibtex information
        """
        try:
            record = self.oai.getRecord(
                metadataPrefix="marcxml", identifier="oai:inspirehep.net:" + inspireID
            )
        except (URLError, ErrorBase, IncompleteRead, SocketError):
            pBLogger.exception("Impossible to get marcxml for entry %s" % inspireID)
            return False
        nhand = 0
        if verbose > 0:
            pBLogger.info("Reading data --- " + time.strftime("%c"))
        try:
            if record[1] is None:
                raise ValueError("Empty record!")
            res = self.readRecord(record[1], readConferenceTitle=readConferenceTitle)
            res["id"] = inspireID
            if bibtex is not None and res["pages"] is not None:
                outcome, bibtex = self.updateBibtex(res, bibtex)
            if verbose > 0:
                pBLogger.info("Done.")
            return res
        except Exception:
            pBLogger.exception("Impossible to read marcxml for entry %s" % inspireID)
            return False

    def updateBibtex(self, res, bibtex):
        """use OAI data to update the (existing) bibtex information
        of an entry

        Parameters:
            res: the recent search results
            bibtex: the old bibtex to be updated

        Output:
            True/False and a string containing the bibtex of the entry
        """
        try:
            element = bibtexparser.loads(bibtex).entries[0]
        except:
            pBLogger.warning("Invalid bibtex!\n%s" % bibtex)
            return False, bibtex
        if res["journal"] is None:
            pBLogger.warning(
                "'journal' from OAI is missing or not a string (recid:%s)" % (res["id"])
            )
            return False, bibtex
        try:
            for k in ["doi", "volume", "pages", "year", "journal"]:
                if res[k] != "" and res[k] is not None:
                    element[k] = res[k]
        except KeyError:
            pBLogger.warning("Something from OAI is missing (recid:%s)" % (res["id"]))
            return False, bibtex
        db = BibDatabase()
        db.entries = [element]
        return True, pbWriter.write(db)

    def retrieveOAIUpdates(self, date1, date2):
        """Harvest the OAI API to get all the updates
        and new occurrences between two dates

        Parameters:
            date1, date2: dates that define
                the time interval to be searched

        Output:
            a list of dictionaries containing the bibtex information
        """
        recs = self.oai.listRecords(
            metadataPrefix="marcxml", from_=date1, until=date2, set="INSPIRE:HEP"
        )
        nhand = 0
        pBLogger.info("\nSTARTING OAI harvester --- " + time.strftime("%c") + "\n\n")
        foundObjects = []
        for count, rec in enumerate(recs):
            id = rec[0].identifier()
            if count % 500 == 0:
                pBLogger.info("Processed %d elements" % count)
            record = rec[1]  # Get pyMARC representation
            if not record:
                continue
            try:
                tmpDict = self.readRecord(record)
                id_ = id.replace("oai:inspirehep.net:", "")
                tmpDict["id"] = id_
                foundObjects.append(tmpDict)
            except Exception as e:
                pBLogger.exception("%s, %s\n%s" % (count, id, e))
        pBLogger.info("Processed %d elements" % count)
        pBLogger.info("END --- " + time.strftime("%c") + "\n\n")
        return foundObjects
